import sys, os

import random
import math

import logging

from . import modelOperationBase

from ...utils.selections import splitStateAccordingStructure
from ...utils.input import getSubParameters
from ...utils.geometry import distributeRandomlyGeneratorChecker

from scipy.spatial.transform import Rotation

import numpy as np

class distributeRandomly(modelOperationBase):

    """
    Component name: distributeRandomly
    Component type: modelOperation

    Author: Pablo Ibáñez-Freire
    Date: 26/08/2023

    Distribution of the selected particles in different bounds

    :param mode: Mode of distribution, the bound where the particles will be distributed, can be "box" or "sphere"
    :type mode: dict, optional, default='box'
    :param avoidClashes: If is larger than 0, the particles will be distributed avoiding clashes with other particles.
                         The value of this parameter is the number of tries to avoid the clash.
                         If the number of tries is reached, an error will be raised.
    :type avoidClashes: int, optional, default=0
    :param randomRotation: If True, the particles will be randomly rotated before being distributed
    :type randomRotation: bool, optional, default=True

    """

    def __randomBoxPoint(self):
        mp = np.random.uniform(low  = [-self.boxX,-self.boxY,-self.boxZ],
                               high = [ self.boxX, self.boxY, self.boxZ])
        return mp

    def __boxCheckDistance(self,p1,p2,r1,r2):
        # Check if the distance between two points is larger than the sum of the radius
        # of the particles, taking into account the periodicity of the box

        # Calculate the distance between the two points
        dr = p1-p2
        # Apply mic periodicity
        dr[0] = dr[0] - math.floor(dr[0]/self.box[0]+0.5)*self.box[0]
        dr[1] = dr[1] - math.floor(dr[1]/self.box[1]+0.5)*self.box[1]
        dr[2] = dr[2] - math.floor(dr[2]/self.box[2]+0.5)*self.box[2]

        d = np.linalg.norm(dr)

        # Check if the distance is larger than the sum of the radius
        if d > 1.05*(r1+r2):
            return True

        return False

    def __randomSpherePoint(self):
        # Generate a random point in a sphere
        # of center self.center and radius self.radius

        maxRad = np.max([np.max(r) for r in self.modelsRads])

        # Generate random radius
        rho = (self.radius-maxRad)*(random.random() ** (1/3))

        # Generate random angles
        theta = math.acos(2 * random.random() - 1)  # Polar angle
        phi = 2 * math.pi * random.random()         # Azimuthal angle

        # Convert to Cartesian coordinates
        x = rho * math.sin(theta) * math.cos(phi)
        y = rho * math.sin(theta) * math.sin(phi)
        z = rho * math.cos(theta)

        mp = np.array([x,y,z]) + self.center

        return mp

    def __sphereCheckDistance(self,p1,p2,r1,r2):
        # Check if the distance between two points is larger than the sum of the radius
        # of the particles

        # Calculate the distance between the two points
        d = np.linalg.norm(p1-p2)

        # Check if the distance is larger than the sum of the radius
        if d > 1.05*(r1+r2):
            return True

        return False

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"mode","avoidClashes","randomRotation"},
                         requiredParameters  = set(),
                         availableSelections = {"selection"},
                         requiredSelections  = {"selection"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        self.randomRotation = params.get("randomRotation",True)

        self.mode,self.modeParams = getSubParameters("mode",params)

        if self.mode == "box" or self.mode == None:
            self.box  = self.getEnsemble().getEnsembleComponent("box")
            self.boxX,self.boxY,self.boxZ = [b/2.0 for b in self.box]

            self.getRandomPoint = self.__randomBoxPoint
            self.checkDistance  = self.__boxCheckDistance

        elif self.mode == "sphere":

            requiredModeParams = {"center","radius"}

            if not requiredModeParams.issubset(self.modeParams):
                self.logger.error(f"Missing parameters for mode {self.mode}: {requiredModeParams}")
                raise Exception("Missing parameters")

            self.center = self.modeParams["center"]
            self.radius = self.modeParams["radius"]

            self.box  = self.getEnsemble().getEnsembleComponent("box")
            boxX,boxY,boxZ = [b/2.0 for b in self.box]

            # Check center is inside the box
            if self.center[0] < -boxX or self.center[0] > boxX:
                self.logger.error(f"Center {self.center} is outside the box in X")
                raise Exception("Center outside box")
            if self.center[1] < -boxY or self.center[1] > boxY:
                self.logger.error(f"Center {self.center} is outside the box in Y")
                raise Exception("Center outside box")
            if self.center[2] < -boxZ or self.center[2] > boxZ:
                self.logger.error(f"Center {self.center} is outside the box in Z")
                raise Exception("Center outside box")

            # Check sphere is inside the box
            if self.center[0] - self.radius < -boxX or self.center[0] + self.radius > boxX:
                self.logger.error(f"Sphere is outside the box in X")
                raise Exception("Sphere outside box")
            if self.center[1] - self.radius < -boxY or self.center[1] + self.radius > boxY:
                self.logger.error(f"Sphere is outside the box in Y")
                raise Exception("Sphere outside box")
            if self.center[2] - self.radius < -boxZ or self.center[2] + self.radius > boxZ:
                self.logger.error(f"Sphere is outside the box in Z")
                raise Exception("Sphere outside box")

            self.getRandomPoint = self.__randomSpherePoint
            self.checkDistance  = self.__sphereCheckDistance

        else:
            self.logger.error(f"Mode {self.mode} not recognized, available modes are 'box' and 'sphere'")
            raise Exception(f"Mode not recognized")

        avoidClashes = params.get("avoidClashes",0)

        ############################################################

        selectedIds = self.getSelection("selection")

        mods = np.asarray(self.getIdsStructure(selectedIds,"modelId"))
        pos  = np.asarray(self.getIdsState(selectedIds,"position"))
        rads = self.getIdsProperty(selectedIds,"radius")

        self.modelsPos = splitStateAccordingStructure(state=pos,
                                                      structure=mods)

        self.modelsRads = splitStateAccordingStructure(state=rads,
                                                       structure=mods)

        if self.randomRotation:
            for i in range(len(self.modelsPos)):

                mp = self.modelsPos[i]

                center = np.mean(mp,axis=0)
                mp -= center

                #Use scipy to generate a random rotation matrix
                R = Rotation.random().as_matrix()

                mp = np.dot(mp,R.T)

                mp += center

                self.modelsPos[i] = mp

        if not avoidClashes:
            newPositions = []
            for i in range(len(self.modelsPos)):
                center = np.mean(self.modelsPos[i],axis=0)
                # Generete a random 3D position
                mp = self.getRandomPoint()

                transVec = mp - center

                newPositions.append([list(p+transVec) for p in self.modelsPos[i]])

            newPositions = [p for mp in newPositions for p in mp]
        else:
            newPositions = distributeRandomlyGeneratorChecker(self.box,
                                                              self.modelsPos,self.modelsRads,
                                                              self.getRandomPoint,self.checkDistance,
                                                              avoidClashes,
                                                              1.05)

        self.setIdsState(selectedIds,"position",newPositions)




from VLMP.components.modelExtensions import modelExtensionBase

class surface(modelExtensionBase):

    """
    Component name: surface
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 17/06/2023

    Common epsilon, sigma surface for particles in the system.

    :param epsilon: epsilon of the surface
    :type epsilon: float
    :param surfacePosition: position of the surface
    :type surfacePosition: float
    :param ignoredTypes: types to ignore
    :type ignoredTypes: list

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"epsilon",
                                                "surfacePosition"},
                         requiredParameters  = {"surfacePosition"},
                         availableSelections = {"selection"},
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        surfacePosition = params.get("surfacePosition",0.0)
        epsilon         = params.get("epsilon",1.0)

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["Surface","SurfaceGeneralLennardJonesType2"]
        extension[name]["parameters"] = {"surfacePosition":surfacePosition}

        extension[name]["labels"] = ["name","epsilon","sigma"]
        extension[name]["data"] = []

        types = self.getTypes()
        for typ,info in types.getTypes().items():
            extension[name]["data"].append([typ,epsilon,info["radius"]])

        ############################################################

        self.setExtension(extension)



import sys, os

import logging

from . import simulationStepBase

class patchPolymersMeasurement(simulationStepBase):
    """
    Component name: patchPolymers
    Component type: simulationStep

    Author: Pablo Ibáñez-Freire
    Date: 7/05/2023

    This step is used to measure properties of the
    polymers created by dynamic bonded patchy particles.
    It computes size of the polymers and if they are
    bonded to the surface or not.

    :param startStep: First step to apply the simulationStep
    :type startStep: int, optional
    :param endStep: Last step to apply the simulationStep
    :type endStep: int, optional

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"outputFilePath","bufferSize","surfaceEnergyThreshold"},
                         requiredParameters  = {"outputFilePath","bufferSize","surfaceEnergyThreshold"},
                         availableSelections = {"selection"},
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        #Check if intervalStep is different from 1. If so, raise an error
        if params.get("intervalStep") != 1:
            self.logger.error("[PatchPolymersMeasurement] intervalStep must be 1")
            raise Exception("Not valid intervalStep")

        parameters = {}

        parameters["outputFilePath"]         = params["outputFilePath"]
        parameters["bufferSize"]             = params["bufferSize"]
        parameters["surfaceEnergyThreshold"] = params["surfaceEnergyThreshold"]

        simulationStep = {
            name:{
              "type":["TopologicalMeasures","PatchPolymers"],
              "parameters":{**parameters}
            }
        }

        ############################################################

        self.setGroup("selection")
        self.setSimulationStep(simulationStep)






from VLMP.components.models import modelBase

import json
import copy

class FILE(modelBase):
    """
    Component name: FILE
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 17/06/2023

    Load model from file

    :param inputFilePath: Path to the input file
    :type inputFilePath: str
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"inputFilePath"},
                         requiredParameters  = {"inputFilePath"},
                         definedSelections   = {"particleId","forceField"},
                         **params)

        ############################################################

        self.inputFilePath = params["inputFilePath"]
        self.logger.info(f"[FILE] Loading model from file {self.inputFilePath}")

        ########################################################

        with open(self.inputFilePath) as f:
            inputJSON = json.load(f)

        # TYPES

        types = self.getTypes()

        typesLabels = inputJSON["global"]["types"]["labels"]
        for typ in inputJSON["global"]["types"]["data"]:
            typInfo = {l:typ[i] for i,l in enumerate(typesLabels)}
            types.addType(**typInfo)

        #Generate positions, a line along the z axis
        state = copy.deepcopy(inputJSON["state"])

        #Generate structure
        structure = copy.deepcopy(inputJSON["topology"]["structure"])

        #Generate forceField
        forceField = copy.deepcopy(inputJSON["topology"]["forceField"])

        ########################################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)


    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += params["particleId"]

        if "forceField" in params:
            sel += self.getForceFieldSelection(params["forceField"])

        return sel



import sys, os

import logging

import numpy as np

from . import modelExtensionBase

class constantTorqueBetweenCentersOfMass(modelExtensionBase):

    """
    Component name: constantTorqueBetweenCentersOfMass
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 04/04/2023

    Constant torque between centers of mass of selected particles

    :param selection1: Selection for the first particle group
    :type selection1: list of dictionaries
    :param selection2: Selection for the second particle group
    :type selection2: list of dictionaries
    :param torque: torque applied to the particles
    :type torque: float

    ...
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"torque"},
                         requiredParameters  = {"torque"},
                         availableSelections = {"selection1","selection2"},
                         requiredSelections  = {"selection1","selection2"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        torque = params.get("torque")
        #Check it torque is a float
        if not isinstance(torque,float):
            raise Exception("torque must be a float")

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["Set2","ConstantTorqueBetweenCentersOfMass"]
        extension[name]["parameters"] = {}
        extension[name]["labels"] = ["idSet_i","idSet_j","torque"]
        extension[name]["data"]   = []

        selectedIds1 = self.getSelection("selection1")
        selectedIds2 = self.getSelection("selection2")

        extension[name]["data"].append([selectedIds1,selectedIds2,torque])

        ############################################################

        self.setExtension(extension)






import sys, os

import logging

from . import systemBase

class backup(systemBase):
    """
    Component name: backup
    Component type: system

    Author: Pablo Ibáñez-Freire
    Date: 13/03/2023

    Component used to add a backup system to the simulation.
    When this component is added to the simulation, the simulation will
    create a backup of the simulation every backupIntervalStep steps.
    The simulation will try to restore the simulation from the backup
    if the simulation crashes.

    :param backupIntervalStep: Interval of steps between backups
    :type backupIntervalStep: int
    :param backupFilePath: Path to the backup file, defaults to "backup"
    :type simulationName: str, optional
    :param backupStartStep: Step at which the backup starts
    :type backupStartStep: int, optional
    :param backupEndStep: Step at which the backup ends
    :type backupEndStep: int, optional
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters  = {"backupIntervalStep",
                                                 "backupStartStep","backupEndStep",
                                                 "backupFilePath"},
                         requiredParameters = {"backupIntervalStep"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        system = {
            name:{"type":["Simulation","Backup"],
                  "parameters":{}}
        }

        system[name]["parameters"]["backupIntervalStep"] = params.get("backupIntervalStep")
        system[name]["parameters"]["backupFilePath"]     = params.get("backupFilePath","backup")

        if "backupStartStep" in params:
            system[name]["parameters"]["backupStartStep"] = params.get("backupStartStep")
        if "backupEndStep" in params:
            system[name]["backupEndStep"] = params.get("backupEndStep")

        ############################################################

        self.setSystem(system)










import sys, os
import copy

import logging

from . import ensembleBase

class NVTlambda(ensembleBase):

    """
    Component name: NVTlambda
    Component type: ensemble

    Author: Pablo Ibáñez-Freire
    Date: 6/11/2023

    Set the box size, temperature and lambda. Lambda is the parameter for thermodynamic integration.

    :param box: box size
    :type box: float
    :param temperature: temperature
    :type temperature: float
    :param lambda: lambda
    :type lambda: float

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"box","temperature","lambda"},
                         requiredParameters  = {"box","temperature","lambda"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        box         = params.get("box")
        temperature = params.get("temperature")
        lambda_     = params.get("lambda")

        self.setEnsembleName("NVTlambda")

        self.addEnsembleComponent("box",box)
        self.addEnsembleComponent("temperature",temperature)
        self.addEnsembleComponent("lambda",lambda_)



import sys, os
import json

import copy
import itertools

import logging

import random

from scipy.spatial import cKDTree

import numpy as np
import orthopoly

from . import modelBase

from ...utils.input import getLabelIndex

from ...utils.geometry import quaternionFromVectors
from ...utils.geometry import getEz

from scipy.spatial.transform import Rotation as R

class CORONAVIRUS(modelBase):
    """
    Component name: CORONAVIRUS
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 4/04/2023

    Coronoavirus model

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"nLipids",
                                                "lipidRadius",
                                                "vesicleRadius",
                                                "center",
                                                "nSpikes",
                                                "epsilonLipids_kT",
                                                "muLipids","chiLipids",
                                                "thetaLipids",
                                                "epsilonLipidLipidWithProtein_kT",
                                                "muLipidLipidWithProtein","chiLipidLipidWithProtein",
                                                "thetaLipidLipidWithProtein",
                                                "epsilonLipidWithProtein_kT",
                                                "muLipidWithProtein","chiLipidWithProtein",
                                                "thetaLipidWithProtein",
                                                "proteinProteinEpsilon_kT",
                                                "proteinLipidEpsilon_kT",
                                                "proteinSurfaceEpsilon_kT",
                                                "proteinPeakSurfaceEpsilon_kT",
                                                "peakProteins",
                                                "lipidSurfaceEpsilon_kT",
                                                "surface",
                                                "surfacePosition",
                                                "inputModelData"},
                         requiredParameters  = set(),
                         definedSelections   = {"particleId","type"},
                         **params)

        self.logger = logging.getLogger("VLMP")

        ############################################################
        ############################################################
        ############################################################

        units = self.getUnits()

        if units.getName() != "KcalMol_A":
            self.logger.error(f"[CORONAVIRUS] Units are not set correctly. "
                              f"Please set units to \"KcalMol_A\" (selected: {units.getName()})")
            raise Exception("Not correct units")

        ########################################################

        self.nLipids     = params.get("nLipids",1501)
        self.lipidRadius = params.get("lipidRadius",18.0)
        self.vesRadius   = params.get("vesicleRadius",400.0)

        self.center      = params.get("center",[0.0,0.0,0.0])

        self.nSpikes     = params.get("nSpikes",0)

        self.epsilonLipids = params.get("epsilonLipids_kT",5.0)
        self.muLipids      = params.get("muLipids",3.0)
        self.chiLipids     = params.get("chiLipids",7.0)
        self.thetaLipids   = params.get("thetaLipids",0.0)

        self.epsilonLipidLipidWithProtein = params.get("epsilonLipidLipidWithProtein_kT",self.epsilonLipids)
        self.muLipidLipidWithProtein      = params.get("muLipidLipidWithProtein",self.muLipids)
        self.chiLipidLipidWithProtein     = params.get("chiLipidLipidWithProtein",self.chiLipids)
        self.thetaLipidLipidWithProtein   = params.get("thetaLipidLipidWithProtein",self.thetaLipids)

        self.epsilonLipidWithProtein = params.get("epsilonLipidWithProtein_kT",self.epsilonLipids)
        self.muLipidWithProtein      = params.get("muLipidWithProtein",self.muLipids)
        self.chiLipidWithProtein     = params.get("chiLipidWithProtein",self.chiLipids)
        self.thetaLipidWithProtein   = params.get("thetaLipidWithProtein",self.thetaLipids)

        self.proteinProteinEpsilon = params.get("proteinProteinEpsilon_kT",1.0)
        self.proteinLipidEpsilon   = params.get("proteinLipidEpsilon_kT",1.0)

        self.proteinSurfaceEpsilon     = params.get("proteinSurfaceEpsilon_kT",1.0)
        self.proteinPeakSurfaceEpsilon = params.get("proteinPeakSurfaceEpsilon_kT",self.proteinSurfaceEpsilon)

        self.peakProteins = params.get("peakProtein",["S10"])

        self.lipidSurfaceEpsilon = params.get("lipidSurfaceEpsilon_kT",1.0)

        self.epsilonLipids                *= units.getConstant("kT")
        self.epsilonLipidLipidWithProtein *= units.getConstant("kT")
        self.epsilonLipidWithProtein      *= units.getConstant("kT")
        self.proteinProteinEpsilon        *= units.getConstant("kT")
        self.proteinLipidEpsilon          *= units.getConstant("kT")
        self.proteinSurfaceEpsilon        *= units.getConstant("kT")
        self.proteinPeakSurfaceEpsilon    *= units.getConstant("kT")
        self.lipidSurfaceEpsilon          *= units.getConstant("kT")

        self.surface = params.get("surface",False)
        if self.surface:
            self.surfacePosition = params.get("surfacePosition",0.0)
        if "surfacePosition" in params and not self.surface:
            self.logger.error(f"[CORONAVIRUS] surfacePosition is set but surface is not set to True")
            raise Exception("Error processing input parameters")

        self.modelData = params.get("inputModelData","./data/CORONAVIRUS.json")
        self.modelData = os.path.join(os.path.dirname(os.path.realpath(__file__)),self.modelData)

        with open(self.modelData,"r") as f:
            self.model = json.load(f)

        ########################################################

        self.__generateCoordinatesAndTopology()

    def processSelection(self,**params):

        sel = set()

        if "particleId" in params:
            sel.add(params["particleId"])

        if "type" in params:
            structure = self.getStructure()
            typeIndex = getLabelIndex("type"  ,structure["labels"])

            if params["type"] == "lipids":
                for d in structure["data"]:
                    if d[typeIndex] in ["LV","LP"]:
                        sel.add(d[0])
            elif params["type"] == "proteins":
                for d in structure["data"]:
                    if d[typeIndex] not in ["LV","LP"]:
                        sel.add(d[0])
            else:
                self.logger.error(f"[CORONAVIRUS] Unknown type {params['type']}")
                raise Exception("Unknown type")

        return list(sel)

    def __generateSphere(self,N,radius):

        if not N%2:
            self.logger.error(f"[CORONAVIRUS] The number of lipids must be odd, but is: {N}")
            raise Exception("The number of lipids must be odd")

        theta,phi = orthopoly.spherical_harmonic.grid_fibonacci(N)
        r=np.asarray([radius]*N)

        positions = np.asarray(orthopoly.spherical_harmonic.sph2cart(r, theta, phi)).T

        orientations = []
        for i,pos in enumerate(positions):

            x,y,z = pos

            q = quaternionFromVectors(np.asarray([0.0,0.0,1.0]),pos)
            orientations.append(q)

        self.lipidsIds   = np.arange(0,N)
        self.lipidsTypes = np.full((N,1),"LV")

        return positions,np.asarray(orientations)

    def __getProteinMaxRadius(self):
        # Load spike positions
        pTypes = self.spikeModel["particleTypes"]

        radiusIndex = getLabelIndex("radius",pTypes["labels"])
        radii = np.asarray([x[radiusIndex] for x in pTypes["data"]])

        r = np.amax(radii)

        self.logger.debug(f"[CORONAVIRUS] Max radius: {r}")

        return r

    def __addSpike(self):
        #Last spike bead is removed, it is considred to be a lipid

        added = False

        spikeCoordTemplate     = self.spikeModel["coordinates"]
        spikeStructureTemplate = self.spikeModel["structure"]

        spikeBondsTemplate       = self.spikeModel["bonds"]
        spikeProtAnglesTemplate  = self.spikeModel["anglesProt"]
        spikeLipidAnglesTemplate = self.spikeModel["anglesProtLipid"]

        offSet = self.lipidsPositions.shape[0]+self.spikePositions.shape[0]

        #Add positions
        while(not added):
            self.logger.debug(f"[CORONAVIRUS] Trying to add the spike {self.addedSpikes+1}")

            # Select random lipid
            lipidId = random.randint(0,self.nLipids-1)

            # Load spike positions
            posIndex = getLabelIndex("positions",spikeCoordTemplate["labels"])
            currentSpikePositions = np.asarray([x[posIndex] for x in spikeCoordTemplate["data"]])

            # Aling spike
            currentSpikePositions = currentSpikePositions-currentSpikePositions[-1]

            v1 = currentSpikePositions[-2]-currentSpikePositions[-1]
            v2 = getEz(self.lipidsOrientations[lipidId])

            q0,q1,q2,q3=quaternionFromVectors(v1,v2)
            rot = R.from_quat(np.asarray([q1,q2,q3,q0]))

            currentSpikePositions = rot.apply(currentSpikePositions)
            currentSpikePositions = currentSpikePositions+self.lipidsPositions[lipidId]

            # Check if spike clash

            if self.spikePositions.any():
                minDst,minDstIndex = cKDTree(self.spikePositions).query(currentSpikePositions, 1)
                minDst = np.amin(minDst)
            else:
                minDst = np.inf

            if minDst > 1.5*(2.0*self.proteinMaxRadius):
                #Remove last
                currentSpikePositions = currentSpikePositions[:-1]

                self.spikePositions   = np.append(self.spikePositions,currentSpikePositions,axis=0)

                orientation = quaternionFromVectors(np.asarray([0.0,0.0,1.0]),currentSpikePositions[-1])
                orientation = orientation*np.ones((currentSpikePositions.shape[0],1))

                self.spikeOrientations = np.append(self.spikeOrientations,orientation,axis=0)
                added = True

        ##Add ids
        idIndex   = getLabelIndex("id"  ,spikeStructureTemplate["labels"])
        typeIndex = getLabelIndex("type",spikeStructureTemplate["labels"])

        currentIds   = [x[idIndex]   for x in spikeStructureTemplate["data"]]
        currentTypes = [x[typeIndex] for x in spikeStructureTemplate["data"]]
        #Remove last
        spikeBaseBeadId   = currentIds[-1]
        spikeBaseBeadType = currentTypes[-1]

        currentIds      = currentIds[:-1]
        currentTypes    = currentTypes[:-1]

        self.spikeIds   = np.append(self.spikeIds,np.asarray(currentIds).T.reshape(-1,1)+offSet,axis=0)
        self.spikeTypes = np.append(self.spikeTypes,np.asarray(currentTypes).T.reshape(-1,1),axis=0)

        currentMdls  = [self.addedSpikes+1]*len(currentIds)
        self.spikeMdls  = np.append(self.spikeMdls,np.asarray(currentMdls).T.reshape(-1,1),axis=0)
        #Change lipid type
        self.lipidsTypes[lipidId]="LP"

        ###########################################

        id_iIndex = getLabelIndex("id_i",spikeBondsTemplate["labels"])
        id_jIndex = getLabelIndex("id_j",spikeBondsTemplate["labels"])
        r0Index   = getLabelIndex("r0"  ,spikeBondsTemplate["labels"])

        for bnd in spikeBondsTemplate["data"]:
            id_i = bnd[id_iIndex]
            id_j = bnd[id_jIndex]
            r0   = bnd[r0Index]

            if id_i == spikeBaseBeadId:
                id_i = lipidId
            else:
                id_i += offSet

            if id_j == spikeBaseBeadId:
                id_j = lipidId
            else:
                id_j += offSet

            self.bonds.append([id_i,id_j,r0])

        ###########################################

        id_iIndex = getLabelIndex("id_i",spikeProtAnglesTemplate["labels"])
        id_jIndex = getLabelIndex("id_j",spikeProtAnglesTemplate["labels"])
        id_kIndex = getLabelIndex("id_k",spikeProtAnglesTemplate["labels"])
        nameIndex = getLabelIndex("name",spikeProtAnglesTemplate["labels"])

        for bnd in spikeProtAnglesTemplate["data"]:
            id_i = bnd[id_iIndex]
            id_j = bnd[id_jIndex]
            id_k = bnd[id_kIndex]
            name = bnd[nameIndex]

            if id_i == spikeBaseBeadId:
                id_i = lipidId
            else:
                id_i += offSet

            if id_j == spikeBaseBeadId:
                id_j = lipidId
            else:
                id_j += offSet

            if id_k == spikeBaseBeadId:
                id_k = lipidId
            else:
                id_k += offSet

            self.angProt.append([id_i,id_j,id_k,name])

        ###########################################

        id_iIndex = getLabelIndex("id_i",spikeLipidAnglesTemplate["labels"])
        id_jIndex = getLabelIndex("id_j",spikeLipidAnglesTemplate["labels"])
        nameIndex = getLabelIndex("name",spikeLipidAnglesTemplate["labels"])
        refIndex  = getLabelIndex("reference",spikeLipidAnglesTemplate["labels"])

        for bnd in spikeLipidAnglesTemplate["data"]:
            id_i = bnd[id_iIndex]
            id_j = bnd[id_jIndex]
            name = bnd[nameIndex]
            ref  = bnd[refIndex]

            if id_i == spikeBaseBeadId:
                id_i = lipidId
            else:
                id_i += offSet

            if id_j == spikeBaseBeadId:
                id_j = lipidId
            else:
                id_j += offSet

            if ref == spikeBaseBeadId:
                ref = lipidId
            else:
                ref += offSet

            self.angLipidProt.append([id_i,id_j,name,ref])

        ###########################################

        self.logger.debug(f"[CORONAVIRUS] Added spike {self.addedSpikes+1}")
        self.addedSpikes+=1

    def __addSpikes(self):
        for n in range(self.nSpikes):
            self.logger.debug(f"[CORONAVIRUS] Adding spike {n+1}/{self.nSpikes}")
            self.__addSpike()

        self.logger.debug(f"[CORONAVIRUS] SpikeIds shape:{self.spikeIds.shape}")
        self.logger.debug(f"[CORONAVIRUS] SpikeTypes shape:{self.spikeTypes.shape}")
        self.logger.debug(f"[CORONAVIRUS] SpikePositions shape:{self.spikePositions.shape}")
        self.logger.debug(f"[CORONAVIRUS] SpikeOrientations shape:{self.spikeOrientations.shape}")

    def __generateExclusions(self):

        excl = {}

        for bnd in self.bonds:
            i,j,_ = bnd
            excl[i]=[]
            excl[j]=[]
        for ang in self.angProt:
            i,j,k,_ = ang
            excl[i]=[]
            excl[j]=[]
            excl[k]=[]
        for ang in self.angLipidProt:
            i,j,_,_ = ang
            excl[i]=[]
            excl[j]=[]

        for bnd in self.bonds:
            i,j,_ = bnd
            excl[i].append(j)
            excl[j].append(i)
        for ang in self.angProt:
            i,j,k,_ = ang
            excl[i].append(j)
            excl[i].append(k)

            excl[j].append(i)
            excl[j].append(k)

            excl[k].append(i)
            excl[k].append(j)
        for ang in self.angLipidProt:
            i,j,_,_ = ang
            excl[i].append(j)
            excl[j].append(i)

        self.exclusions = {i:sorted(list(set(data))) for i,data in excl.items()}


    def __generateCoordinatesAndTopology(self):

        ########################################################
        ############# COORDINATES GENERATION STARTS ############
        ########################################################

        self.logger.info(f"[CORONAVIRUS] Generating CORONAVIRUS model with {self.nLipids} lipids "
                         f"and vesicle radius of {self.vesRadius} with {self.nSpikes} spikes")

        self.lipidsPositions,self.lipidsOrientations = self.__generateSphere(self.nLipids,
                                                                             self.vesRadius)

        #Spikes
        self.spikeModel = self.model["spike"]

        #Generate bonds with enm

        self.spikeModel["bonds"] = {}
        self.spikeModel["bonds"]["labels"] = ["id_i", "id_j", "r0"]
        self.spikeModel["bonds"]["data"]   = []

        id_index  = getLabelIndex("id"       ,self.spikeModel["coordinates"]["labels"])
        pos_index = getLabelIndex("positions",self.spikeModel["coordinates"]["labels"])

        coord = self.spikeModel["coordinates"]["data"]

        ##############################################################

        for d in self.model["spike"]["anglesProt"]["data"]:
            if d[3] == "knee":
                link1,link2,link3 = d[0],d[1],d[2]
                avoidIds = [link1,link2]
                break

        for c in coord:
            id_ = int(c[id_index][1::])
            if id_ == link1:
                link1pos = np.asarray(c[pos_index])
            if id_ == link2:
                link2pos = np.asarray(c[pos_index])
            if id_ == link3:
                link3pos = np.asarray(c[pos_index])

        self.spikeModel["bonds"]["data"].append([link1,link2,np.linalg.norm(link1pos-link2pos)])
        self.spikeModel["bonds"]["data"].append([link2,link3,np.linalg.norm(link2pos-link3pos)])

        ##############################################################

        for c1,c2 in itertools.combinations(coord,r=2):
            if(c1[id_index]==c2[id_index]):
                continue

            id1 = int(c1[id_index][1::])
            id2 = int(c2[id_index][1::])

            if id1 in avoidIds:
                continue
            if id2 in avoidIds:
                continue

            p1 = np.asarray(c1[pos_index])
            p2 = np.asarray(c2[pos_index])

            r0 = np.linalg.norm(p1-p2)

            if r0 < self.model["spike"]["parameters"]["bonds"]["bondsProtProt"]["cutOff"]:
                self.spikeModel["bonds"]["data"].append([id1,id2,r0])

        #########################

        self.proteinMaxRadius = self.__getProteinMaxRadius()

        self.addedSpikes  = 0

        self.spikeIds          = np.empty((0,1),dtype=int)
        self.spikeTypes        = np.empty((0,1),dtype=int)
        self.spikeMdls         = np.empty((0,1),dtype=int)
        self.spikePositions    = np.empty((0,3))
        self.spikeOrientations = np.empty((0,4))

        self.bonds        = []
        self.angLipidProt = []
        self.angProt      = []

        self.__addSpikes()

        #Move to center
        for p in self.lipidsPositions:
            p[0]=p[0]+self.center[0]
            p[1]=p[1]+self.center[1]
            p[2]=p[2]+self.center[2]
        for p in self.spikePositions:
            p[0]=p[0]+self.center[0]
            p[1]=p[1]+self.center[1]
            p[2]=p[2]+self.center[2]

        ##Compute initial center

        tpy2radius = {}
        for t in ["LV","LP"]:
            tpy2radius[t]=self.lipidRadius
        for t in self.model["spike"]["particleTypes"]["data"]:
            tpy2radius[t[0]]=t[2]

        minPos = float(self.lipidsPositions[0][2]-tpy2radius[self.lipidsTypes[0][0]])
        for [t],p in zip(self.lipidsTypes,self.lipidsPositions):
            v = float(p[2]-tpy2radius[t])
            minPos=min(minPos,v)
        for [t],p in zip(self.spikeTypes,self.spikePositions):
            v = float(p[2]-tpy2radius[t])
            minPos=min(minPos,v)

        if self.surface:
            surfPos = self.surfacePosition
            if(minPos < surfPos):
                trans = surfPos - minPos + 0.01
                for p in self.lipidsPositions:
                    p[2]=p[2]+trans
                for p in self.spikePositions:
                    p[2]=p[2]+trans

        #############################################################
        #############################################################
        #############################################################

        types = self.getTypes()

        tpy2radius = {}
        for t in ["LV","LP"]:
            tpy2radius[t]=self.lipidRadius
            types.addType(name=t,mass=1.0,radius=self.lipidRadius)
        for t in self.model["spike"]["particleTypes"]["data"]:
            tName,tMass,tRadius,tCharge = t
            tpy2radius[t[0]]=tRadius
            types.addType(name=tName,mass=tMass,radius=tRadius,charge=tCharge)

        state = {}

        state = {}
        state["labels"]=["id","position","direction"]
        state["data"]  =[]

        for i,p,q in zip(self.lipidsIds,self.lipidsPositions,self.lipidsOrientations):
            p=np.around(p,2)
            state["data"].append([int(i),list(p),list(q)])
        for i,p,q in zip(self.spikeIds,self.spikePositions,self.spikeOrientations):
            p=np.around(p,2)
            state["data"].append([int(i),list(p),list(q)])

        structure={}
        structure["labels"] = ["id", "type", "modelId"]
        structure["data"]   = []

        id2tpy = {}
        for id,t in zip(self.lipidsIds,self.lipidsTypes):
            id = int(id)
            t  = str(t[0])
            id2tpy[id]=t
            structure["data"].append([id,t,0])
        for id,t,m in zip(self.spikeIds,self.spikeTypes,self.spikeMdls):
            id = int(id)
            t  = str(t[0])
            id2tpy[id]=t
            structure["data"].append([id,t,int(m[0])])

        forceField = {}

        forceField["groups"] = {}
        forceField["groups"]["type"] = ["Groups","GroupsList"]
        forceField["groups"]["parameters"] = {}
        forceField["groups"]["labels"] = ["name", "type", "selection"]
        forceField["groups"]["data"]   = [
            ["pgLipids","Types",["LV","LP"]]
        ]

        self.__generateExclusions()

        forceField["verletList"] = {}
        forceField["verletList"]["type"]       = ["VerletConditionalListSet", "nonExclIntra_nonExclInter"]
        forceField["verletList"]["parameters"] = {"cutOffVerletFactor":1.2}
        forceField["verletList"]["labels"]     = ["id","id_list"]
        forceField["verletList"]["data"]       = []

        for i,excl in self.exclusions.items():
            forceField["verletList"]["data"].append([i,excl])

        #Lipids

        forceField["lipidsLipids"] = {}
        forceField["lipidsLipids"]["type"] = ["NonBonded", "Zhang"]
        forceField["lipidsLipids"]["parameters"] = {"condition":"intra","group":"pgLipids"}
        forceField["lipidsLipids"]["labels"] = ["name_i", "name_j", "radius", "epsilon", "mu", "chi", "theta"]
        forceField["lipidsLipids"]["data"]   = [
            ["LV","LV",self.lipidRadius,self.epsilonLipids,self.muLipids,self.chiLipids,self.thetaLipids],
            ["LV","LP",self.lipidRadius,self.epsilonLipidLipidWithProtein,self.muLipidLipidWithProtein,self.chiLipidLipidWithProtein,self.thetaLipidLipidWithProtein],
            ["LP","LV",self.lipidRadius,self.epsilonLipidLipidWithProtein,self.muLipidLipidWithProtein,self.chiLipidLipidWithProtein,self.thetaLipidLipidWithProtein],
            ["LP","LP",self.lipidRadius,self.epsilonLipidWithProtein,self.muLipidWithProtein,self.chiLipidWithProtein,self.thetaLipidWithProtein]
        ]

        forceField["nonPolar"] = {}
        forceField["nonPolar"]["type"] = ["NonBonded", "GeneralLennardJonesType2"]
        forceField["nonPolar"]["parameters"] = {"cutOffFactor": 2.5,"condition":"inter"}
        forceField["nonPolar"]["labels"] = ["name_i", "name_j", "epsilon", "sigma"]
        forceField["nonPolar"]["data"]   = []

        for t1 in tpy2radius.keys():
            for t2 in tpy2radius.keys():
                is1lipid = t1 in ["LV","LP"]
                is2lipid = t2 in ["LV","LP"]

                r1 = tpy2radius[t1]
                r2 = tpy2radius[t2]

                if is1lipid and is2lipid:
                    forceField["nonPolar"]["data"].append([t1,t2,0.0,0.0])
                if is1lipid and not is2lipid:
                    forceField["nonPolar"]["data"].append([t1,t2,self.proteinLipidEpsilon,r1+r2])
                if not is1lipid and is2lipid:
                    forceField["nonPolar"]["data"].append([t1,t2,self.proteinLipidEpsilon,r1+r2])
                if not is1lipid and not is2lipid:
                    forceField["nonPolar"]["data"].append([t1,t2,self.proteinProteinEpsilon,r1+r2])

        #Bonds
        if self.nSpikes > 0:

            Kpp = self.model["spike"]["parameters"]["bonds"]["bondsProtProt"]["K"]
            Kpl = self.model["spike"]["parameters"]["bonds"]["bondsLipidsProt"]["K"]

            forceField["bonds"] = {}
            forceField["bonds"]["type"]       = ["Bond2","Harmonic"]
            forceField["bonds"]["labels"]     = ["id_i","id_j","r0","K"]
            forceField["bonds"]["parameters"] = {}
            forceField["bonds"]["data"]       = []

            for i,j,r0 in self.bonds:
                if "L" in id2tpy[i] or "L" in id2tpy[j]:
                    K=Kpl
                else:
                    K=Kpp
                forceField["bonds"]["data"].append([i,j,r0,K])

            #angles
            angParam = self.model["spike"]["parameters"]["angles"]["anglesProt"]

            forceField["angProt"] = {}
            forceField["angProt"]["type"]       = ["Bond3","HarmonicAngular"]
            forceField["angProt"]["labels"]     = ["id_i","id_j","id_k","theta0","K"]
            forceField["angProt"]["parameters"] = {}
            forceField["angProt"]["data"]       = []

            for i,j,k,name in self.angProt:
                theta0 = angParam[name]["theta"]
                K      = angParam[name]["K"]
                forceField["angProt"]["data"].append([i,j,k,np.pi-np.radians(theta0),K])

            ########

            #angParam = self.model["spike"]["parameters"]["angles"]["anglesProtLipid"]

            #forceField["angLipidProt"] = {}
            #forceField["angLipidProt"]["type"]       = ["Bond2","Alignment"]
            #forceField["angLipidProt"]["labels"]     = ["id_i","id_j","theta0","K","reference"]
            #forceField["angLipidProt"]["parameters"] = {}
            #forceField["angLipidProt"]["data"]       = []

            #for i,j,name,ref in self.angLipidProt:
            #    theta0 = angParam[name]["theta"]
            #    K      = angParam[name]["K"]
            #    forceField["angLipidProt"]["data"].append([i,j,np.pi-np.radians(theta0),K,ref])

        if self.surface:
            forceField["surface"] = {}
            forceField["surface"]["type"]       = ["Surface", "SurfaceGeneralLennardJonesType2"]
            forceField["surface"]["parameters"] = {"surfacePosition":self.surfacePosition}
            forceField["surface"]["labels"] = ["name","epsilon","sigma"]
            forceField["surface"]["data"]   = []

            for t in tpy2radius.keys():
                isLipid = t in ["LV","LP"]
                r       = tpy2radius[t]

                if isLipid:
                    forceField["surface"]["data"].append([t,self.lipidSurfaceEpsilon,r])
                else:
                    isPeakProtein = t in self.peakProteins
                    if isPeakProtein:
                        forceField["surface"]["data"].append([t,self.proteinPeakSurfaceEpsilon,r])
                    else:
                        forceField["surface"]["data"].append([t,self.proteinSurfaceEpsilon,r])

        #############################################################
        #############################################################
        #############################################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)



import os
import copy

import logging

################ INTEGRATORS INTERFACE ################

from pyUAMMD import simulation

class integratorBase:

    def __init__(self,
                 _type:str,_name:str,
                 units,types,ensemble,models,
                 availableParameters:set,
                 requiredParameters:set,
                 **params):

        self.logger = logging.getLogger("VLMP")

        self._type = _type
        self._name = _name

        self._units    = units
        self._types    = types
        self._ensemble = ensemble
        self._models   = models

        self.availableParameters  = availableParameters.copy()
        self.requiredParameters = requiredParameters.copy()

        # Check all required parameters are available parameters
        if not self.requiredParameters.issubset(self.availableParameters):
            notAvailable = self.requiredParameters.difference(self.availableParameters)
            self.logger.error(f"[Integrator] ({self._type}) Some required parameters ({notAvailable}) are not available parameters for integrator {self._name}")
            raise Exception(f"Required paramaters are not available parameters")

        # Check if all parameters given by params are available
        for par in params:
            if par not in self.availableParameters:
                self.logger.error(f"[Integrator] ({self._type}) Parameter {par} not available for integrator {self._name}")
                raise Exception(f"Parameter not available")

        # Check if all required parameters are given
        for par in self.requiredParameters:
            if par not in params:
                self.logger.error(f"[Integrator] ({self._type}) Required parameter {par} not given for integrator {self._name}")
                raise Exception(f"Required parameter not given")

        self.logger.info(f"[Integrator] ({self._type}) Using integrator {self._name}")

        ########################################################

        self._integrator       = None
        self._integrationSteps = None

    ########################################################

    def getName(self):
        return self._name

    def getType(self):
        return self._type

    ########################################################

    def getUnits(self):
        return self._units

    def getTypes(self):
        return self._types

    def getEnsemble(self):
        return self._ensemble

    ########################################################

    def setIntegrator(self, integrator):
        self._integrator = integrator

    def setIntegrationSteps(self, integrationSteps):
        self._integrationSteps = integrationSteps

    def getIntegrator(self):
        if self._integrator is None:
            self.logger.error(f"[Integrator] ({self._type}) Integrator not initialized")
            raise Exception(f"Integrator not initialized")
        return self._integrator

    def getIntegrationSteps(self):
        if self._integrationSteps is None:
            self.logger.error(f"[Integrator] ({self._type}) Integration steps not initialized")
            raise Exception(f"Integration steps not initialized")
        return self._integrationSteps

    ########################################################

    def getSimulation(self,DEBUG_MODE = False):
        sim = {}

        sim["integrator"] = {self._name:copy.deepcopy(self.getIntegrator())}

        sim["integrator"]["schedule"] = {
            "type":["Schedule","Integrator"],
            "labels":["order","integrator","steps"],
            "data":[[1,self._name,self.getIntegrationSteps()]]
        }

        return simulation(copy.deepcopy(sim),DEBUG_MODE)

############### IMPORT ALL INTEGRATORS ###############

import glob

currentPath = os.path.dirname(os.path.abspath(__file__))
integrators = [os.path.basename(x) for x in glob.glob(currentPath+"/*.py")]
integrators = [x.split(".")[0] for x in integrators if "__" not in x]

for i in integrators:
    try:
        exec(f"from .{i} import *")
    except Exception as e:
        logging.getLogger("VLMP").error(e)
        logging.getLogger("VLMP").error(f"[Integrator] Error importing integrator type component {i}")



import sys, os

import copy

import logging

from . import integratorBase

class MagneticBrownian(integratorBase):

    """
    Component name: MagneticBrownian
    Component type: integrator

    Author: P. Palacios-Alonso
    Date: 17/10/2023

    Simple Euler-Maruyama integrator for brownian dynamics.

    :param integrationSteps: Number of integration steps.
    :type integrationSteps: int
    :param timeStep: Time step of the integrator.
    :type timeStep: float
    :param viscosity: Viscosity of the fluid.
    :type viscosity: float


    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"integrationSteps","timeStep","viscosity",
                                                "gyroRatio", "damping", "msat",
                                                "magneticIntegrationAlgorithm"},
                         requiredParameters  = {"integrationSteps","timeStep","viscosity",
                                                "gyroRatio", "damping", "msat",
                                                "magneticIntegrationAlgorithm"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        integrationSteps = params.get("integrationSteps")

        parameters = {}

        parameters["timeStep"]  = params.get("timeStep")
        parameters["viscosity"] = params.get("viscosity")
        parameters["gyroRatio"] = params.get("gyroRatio")
        parameters["damping"]   = params.get("damping")
        parameters["msat"]      = params.get("msat")
        parameters["magneticIntegrationAlgorithm"] = params.get("magneticIntegrationAlgorithm")

        
        integrator = {
            "type" : ["Magnetic", "Brownian"],
            "parameters" : copy.deepcopy(parameters)
        }

        ############################################################

        self.setIntegrationSteps(integrationSteps)
        self.setIntegrator(integrator)



import sys, os

import logging

from . import typesBase

class none(typesBase):

    """
    Component name: none
    Component type: types

    Author: Pablo Ibáñez-Freire
    Date: 25/04/2023

    None types. No components are added.
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = set(),
                         requiredParameters  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        self.setTypesName("None")



from VLMP.components.modelOperations import modelOperationBase

import numpy as np

from scipy.spatial.transform import Rotation as R

class rotation(modelOperationBase):

    """
    Component name: rotation
    Component type: modelOperation

    Author: Pablo Ibáñez-Freire
    Date: 01/09/2023

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"axis","angle"},
                         requiredParameters  = {"axis","angle"},
                         availableSelections = {"selection"},
                         requiredSelections  = {"selection"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        rotAxis = params["axis"]
        angle   = params["angle"]

        ############################################################

        selectedIds = self.getSelection("selection")

        if len(selectedIds) == 0:
            self.error("No elements selected.")
            raise Exception("No elements selected.")

        if len(selectedIds) > 1:

            pos    = np.asarray(self.getIdsState(selectedIds,"position"))

            center = np.mean(pos,axis=0)
            pos    = pos - center

            rotAxis = np.asarray(rotAxis)
            rotAxis = rotAxis/np.linalg.norm(rotAxis)

            r = R.from_rotvec(angle*rotAxis)
            pos = r.apply(pos)

            pos = pos + center

            self.setIdsState(selectedIds,"position",pos.tolist())



from VLMP.components.modelOperations import modelOperationBase

import numpy as np

class setParticleLowestPosition(modelOperationBase):

    """
    Component name: setParticleLowestPosition
    Component type: modelOperation

    Author: Pablo Ibáñez-Freire
    Date: 17/06/2023

    Set the lowest particle position to value.

    :param position: Position to set the lowest particle to.
    :type position: z coordinate, float
    :param considerRadius: Consider particle radius when setting the lowest position.
    :type considerRadius: bool, optional

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"position","considerRadius","radiusFactor"},
                         requiredParameters  = {"position"},
                         availableSelections = {"selection"},
                         requiredSelections  = {"selection"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        selectedIds = self.getSelection("selection")

        pos    = np.asarray(self.getIdsState(selectedIds,"position"))

        if params.get("considerRadius",False):
            rad = np.asarray(self.getIdsProperty(selectedIds,"radius"))
            radFactor = params.get("radiusFactor",1.0)
            rad = radFactor*rad

        # Find the lowest position
        lowestPosIndex = np.argmin(pos[:,2])
        lowestPos      = pos[lowestPosIndex,2]

        # Set the lowest position to the given value
        offset = 0.0
        if params.get("considerRadius",False):
            rad = np.asarray(self.getIdsProperty(selectedIds,"radius"))
            offset = rad[lowestPosIndex]

        translation = np.asarray([0,0,params["position"] - lowestPos + offset])

        for i in range(len(selectedIds)):
            pos[i] = pos[i] + translation

        self.setIdsState(selectedIds,"position",pos.tolist())



import VLMP
import VLMP.utils.units as unitsUtils

import os

import numpy as np
import matplotlib.pyplot as plt

import json
import jsbeautifier

import copy
import logging

class HighThroughputAFM(VLMP.VLMP):

    def __init__(self,parameters):
        super().__init__()

        requiredAFMParameters = ["K","Kxy",
                                 "epsilon",
                                 "sigma",
                                 "tipVelocity",
                                 "tipMass","tipRadius",
                                 "initialTipSampleDistance"]

        requiredIndentationParameters = ["indentationSteps","thermalizationSteps"]

        self.logger.info("[AFM] Initializing ...")

        for parameter in requiredAFMParameters:
            if parameter not in parameters["AFM"]:
                self.logger.error("[AFM] Required AFM parameter %s not found!" % parameter)
                raise Exception("Required parameter not found!")

        for parameter in requiredIndentationParameters:
            if parameter not in parameters["indentation"]:
                self.logger.error("[AFM] Required indentation parameter %s not found!" % parameter)
                raise Exception("Required parameter not found!")

        self.K   = parameters["AFM"]["K"]
        self.Kxy = parameters["AFM"]["Kxy"]

        self.epsilon = parameters["AFM"]["epsilon"]
        self.sigma   = parameters["AFM"]["sigma"]

        self.tipVelocity     = parameters["AFM"]["tipVelocity"]

        self.tipMass   = parameters["AFM"]["tipMass"]
        self.tipRadius = parameters["AFM"]["tipRadius"]
        self.tipCharge = parameters["AFM"].get("tipCharge",0.0)

        self.initialTipSampleDistance = parameters["AFM"]["initialTipSampleDistance"]
        self.contactResolution        = parameters["AFM"].get("contactResolution",0.1)

        ############################################################################

        self.thermalizationSteps = parameters["indentation"]["thermalizationSteps"]
        self.indentationSteps    = parameters["indentation"]["indentationSteps"]

        if not isinstance(self.thermalizationSteps,list):
            self.thermalizationSteps = [self.thermalizationSteps]
        if not isinstance(self.indentationSteps,list):
            self.indentationSteps = [self.indentationSteps]

        self.fixSampleDuringThermalization = parameters["indentation"].get("fixSampleDuringThermalization",False)
        self.fixSampleDuringIndentation    = parameters["indentation"].get("fixSampleDuringIndentation",False)
        self.KxyFixing                     = parameters["indentation"].get("KxyFixing",None)
        self.backwardIndentation           = parameters["indentation"].get("backwardIndentation",False)

        ############################################################################

        if ( self.fixSampleDuringThermalization or self.fixSampleDuringIndentation ) and self.KxyFixing is None:
            self.logger.error("[AFM] Sample fixing requested, but no KxyFixing provided!")
            raise Exception("KxyFixing not provided!")
        if self.KxyFixing is not None and not ( self.fixSampleDuringThermalization or self.fixSampleDuringIndentation ):
            self.logger.error("[AFM] KxyFixing provided, but no sample fixing requested!")
            raise Exception("No sample fixing requested!")

        if "maxForce" in parameters.keys():
            self.maxForce             = parameters["maxForce"].get("force")
            self.maxForceIntervalStep = parameters["maxForce"].get("maxForceIntervalStep")

            if self.backwardIndentation:
                #Not compatible with maxForce
                self.logger.error("[AFM] Backward indentation requested, but maxForce is not compatible with it!")
                raise Exception("Indentation parameters not compatible!")
        else:
            self.maxForce = None

        if "surface" in parameters.keys():
            self.addSurface = True

            self.epsilonSurface  = parameters["surface"].get("epsilon",1.0)
            self.surfacePosition = parameters["surface"].get("position",0.0)

            self.absorptionHeight = parameters["surface"].get("absorptionHeight",None)

            if self.absorptionHeight is not None:
                self.absorptionK = parameters["surface"].get("absorptionK",None)
                if self.absorptionK is None:
                    self.logger.error("[AFM] Absorption height given but no absorption constant, K!")
                    raise Exception("Absorption K not found!")
                if self.absorptionHeight < self.surfacePosition:
                    self.logger.error("[AFM] Absorption height is below surface!")
                    raise Exception("Absorption height is below surface!")
        else:
            self.addSurface = False

        #Load simulation parameters

        requiredSimulationParameters = ["units","types","temperature","box","samples","integrator"]

        for parameter in requiredSimulationParameters:
            if parameter not in parameters["simulation"]:
                self.logger.error("[AFM] Required parameter %s not found!" % parameter)
                raise Exception("Required parameter not found!")

        self.units  = parameters["simulation"]["units"]
        self.types  = parameters["simulation"]["types"]

        self.temperature = parameters["simulation"]["temperature"]
        #Check if the temperature is a float
        if not isinstance(self.temperature,float):
            raise Exception("The temperature must be a float")

        self.box         = parameters["simulation"]["box"]
        #Check box is a list of 3 floats
        if not isinstance(self.box,list):
            raise Exception("The box must be a list of 3 floats")
        if len(self.box) != 3:
            raise Exception("The box must be a list of 3 floats")
        for i in self.box:
            if not isinstance(i,float):
                raise Exception("The box must be a list of 3 floats")

        self.integrator = parameters["simulation"]["integrator"]

        self.integrator["parameters"]["integrationSteps"] = max(self.thermalizationSteps) + max(self.indentationSteps)
        if self.backwardIndentation:
            # We perform the same number of steps for forward and backward indentation
            self.integrator["parameters"]["integrationSteps"] += max(self.indentationSteps)

        self.samples = parameters["simulation"]["samples"]
        #Check samples is dict of list
        if not isinstance(self.samples,dict):
            raise Exception("The samples must be a dict->dict->list")
        for key in self.samples:
            if not isinstance(self.samples[key],dict):
                raise Exception("The samples must be a dict->dict->list")
            for key2 in self.samples[key]:
                if not isinstance(self.samples[key][key2],list):
                    raise Exception("The samples must be a dict->dict->list")

        nSamples = len(self.samples)
        self.logger.info("[AFM] Number of samples: %d" % nSamples)

        if len(self.thermalizationSteps) != nSamples:
            if len(self.thermalizationSteps) == 1:
                self.logger.info("[AFM] Same thermalization steps for all samples")
                self.thermalizationSteps = [self.thermalizationSteps[0]]*nSamples
            else:
                self.logger.error("[AFM] Different number of thermalization steps and samples!")
                raise Exception("Different number of thermalization steps and samples!")

        if len(self.indentationSteps) != nSamples:
            if len(self.indentationSteps) == 1:
                self.logger.info("[AFM] Same indentation steps for all samples")
                self.indentationSteps = [self.indentationSteps[0]]*nSamples
            else:
                self.logger.error("[AFM] Different number of indentation steps and samples!")
                raise Exception("Different number of indentation steps and samples!")

        self.backupIntervalStep = parameters["simulation"].get("backupIntervalStep",None)

        #Load output parameters
        requiredOutputParameters = []
        for parameter in requiredOutputParameters:
            if parameter not in parameters["output"]:
                self.logger.error("[AFM] Required parameter %s not found!" % parameter)
                raise Exception("Required parameter not found!")

        #Info
        self.infoIntervalStep = parameters["output"].get("infoIntervalStep",None)

        #Save state
        self.saveStateIntervalStep   = parameters["output"].get("saveStateIntervalStep",None)
        self.saveStateOutputFilePath = parameters["output"].get("saveStateOutputFilePath",None)
        self.saveStateOutputFormat   = parameters["output"].get("saveStateOutputFormat",None)

        #If at least one of the save state parameters is specified, check all of them are specified
        self.saveState = False
        if self.saveStateIntervalStep is not None or self.saveStateOutputFilePath is not None or self.saveStateOutputFormat is not None:
            self.saveState = True
            if self.saveStateIntervalStep is None or self.saveStateOutputFilePath is None or self.saveStateOutputFormat is None:
                self.logger.error("[AFM] All the save state parameters must be specified")
                raise Exception("All the save state parameters must be specified")

        #Measurements
        self.afmMeasurementIntervalStep   = parameters["output"].get("afmMeasurementIntervalStep",None)
        self.afmMeasurementOutputFilePath = parameters["output"].get("afmMeasurementOutputFilePath","afm.dat")

        #Extend parameters


        allParameters = ["K", "Kxy",
                         "epsilon",
                         "sigma",
                         "tipVelocity",
                         "tipMass", "tipRadius", "tipCharge",
                         "initialTipSampleDistance",
                         "KxyFixing"]

        if self.addSurface:
            allParameters.extend(["epsilonSurface","surfacePosition"])
            if self.absorptionHeight is not None:
                allParameters.extend(["absorptionHeight","absorptionK"])

        for paramIndex in range(len(allParameters)):
            p = getattr(self,allParameters[paramIndex])
            if isinstance(p,list):
                #Check len matches nSamples
                if len(p) != nSamples:
                    self.logger.error(f"[AFM] The number of samples ({nSamples}) "
                                      f"does not match the number of parameters ({len(p)})"
                                      f" for parameter \"{allParameters[paramIndex]}\"")
                    raise Exception("The number of samples does not match the number of parameters")
            else:
                #Create list of len nSamples
                setattr(self,allParameters[paramIndex],[p]*nSamples)

    def generateSimulationPool(self):


        simulationPool = []

        #Check sample names are unique
        samplesNames = []
        for smp in self.samples.keys():
            if smp in samplesNames:
                self.logger.error("[AFM] Sample name %s is not unique!" % smp)
                raise Exception("Sample name is not unique!")
            else:
                samplesNames.append(smp)

        for index,[smp,smpModels] in enumerate(self.samples.items()):

            sim = {"system":[{"type":"simulationName","parameters":{"simulationName":smp}}],
                   "units":[{"type":self.units}],
                   "types":[{"type":self.types}],
                   "ensemble":[{"type":"NVT","parameters":{"box":self.box,"temperature":self.temperature}}],
                   "integrators":[self.integrator],
                   "models":copy.deepcopy(smpModels.get("models",[])),
                   "modelOperations":copy.deepcopy(smpModels.get("modelOperations",[])),
                   "modelExtensions":copy.deepcopy(smpModels.get("modelExtensions",[])),
                   "simulationSteps":copy.deepcopy(smpModels.get("simulationSteps",[])),
                   }

            thermSteps_smp  = self.thermalizationSteps[index]
            indentSteps_smp = self.indentationSteps[index]

            K_smp   = self.K[index]
            Kxy_smp = self.Kxy[index]

            epsilon_smp = self.epsilon[index]
            sigma_smp   = self.sigma[index]

            tipVelocity_smp = self.tipVelocity[index]

            tipMass_smp   = self.tipMass[index]
            tipRadius_smp = self.tipRadius[index]
            tipCharge_smp = self.tipCharge[index]

            initialTipSampleDistance_smp = self.initialTipSampleDistance[index]
            KxyFixing_smp                = self.KxyFixing[index]

            if self.addSurface:
                epsilonSurface_smp  = self.epsilonSurface[index]
                surfacePosition_smp = self.surfacePosition[index]

                if self.absorptionHeight is not None:
                    absorptionHeight_smp = self.absorptionHeight[index]
                    absorptionK_smp      = self.absorptionK[index]

            #Add tip
            if len(smpModels) > 0:
                sim["models"].append({"name":"TIP","type":"PARTICLE","parameters":{"particleName":"TIP",
                                                                                   "particleMass":tipMass_smp,
                                                                                   "particleRadius":tipRadius_smp,
                                                                                   "particleCharge":tipCharge_smp}})
            else:
                sim["models"].append({"name":"TIP","type":"PARTICLE","parameters":{"particleName":"TIP",
                                                                                   "particleMass":tipMass_smp,
                                                                                   "particleRadius":tipRadius_smp,
                                                                                   "particleCharge":tipCharge_smp,
                                                                                   "position":[0.0,0.0,initialTipSampleDistance_smp+tipRadius_smp]}})

            #Declare selections
            tipSelection = ["TIP"]
            sampleSelection = []

            if len(smpModels) > 0:

                for mdl in smpModels["models"]:
                    if "name" in mdl["parameters"]:
                        sampleSelection.append(mdl["name"])
                    else:
                        sampleSelection.append(mdl["type"])


                ###Model operations
                #if self.maxForce is not None:
                #    sim["modelOperations"].append({"type":"setParticleLowestPosition","parameters":{"position":self.surfacePosition,
                #                                                                                    "considerRadius":True,
                #                                                                                    "radiusFactor":2.0,
                #                                                                                    "selection":{"models":copy.deepcopy(sampleSelection)}}})
                #else:

                sim["modelOperations"].append({"type":"setCenterOfMassPosition","parameters":{"position":[0.0,0.0,0.0],
                                                                                              "selection":{"models":copy.deepcopy(sampleSelection)}}})

                if self.addSurface:
                    sim["modelOperations"].append({"type":"setParticleLowestPosition","parameters":{"position":surfacePosition_smp,
                                                                                                    "considerRadius":True,
                                                                                                    "selection":{"models":copy.deepcopy(sampleSelection)}}})

                sim["modelOperations"].append({"type":"setContactDistance","parameters":{"distance":initialTipSampleDistance_smp,
                                                                                         "resolution":self.contactResolution,
                                                                                         "inverse":True,
                                                                                         "reference":{"models":copy.deepcopy(sampleSelection)},
                                                                                         "mobile":{"models":copy.deepcopy(tipSelection)}}})
            ###Model extensions

            #Add AFM

            sim["modelExtensions"].append({"type":"AFM","parameters":{"K":K_smp,"Kxy":Kxy_smp,
                                                                      "epsilon":epsilon_smp,
                                                                      "sigma":sigma_smp,
                                                                      "tipVelocity":tipVelocity_smp,
                                                                      "indentationStartStep":thermSteps_smp,
                                                                      "tip":{"models":copy.deepcopy(tipSelection)},
                                                                      "sample":{"models":copy.deepcopy(sampleSelection)}}})
            if self.backwardIndentation:
                sim["modelExtensions"][-1]["parameters"]["indentationBackwardStep"] = thermSteps_smp + indentSteps_smp

            fixingStartStep = None
            fixingEndStep   = None
            if       self.fixSampleDuringThermalization and not self.fixSampleDuringIndentation:
                fixingEndStep   = thermSteps_smp
            elif not self.fixSampleDuringThermalization and     self.fixSampleDuringIndentation:
                fixingStartStep = thermSteps_smp
            elif not self.fixSampleDuringThermalization and not self.fixSampleDuringIndentation:
                pass

            #Add sample
            if self.fixSampleDuringThermalization or self.fixSampleDuringIndentation:
                entry = {"type":"constraintCenterOfMassPosition",
                         "parameters":{"K":[KxyFixing_smp,KxyFixing_smp,0.0],
                                       "r0":0.0,
                                       "position":[0.0,0.0,0.0],
                                       "selection":{"models":copy.deepcopy(sampleSelection)}
                                      }
                         }

                if fixingStartStep is not None:
                    entry["parameters"]["startStep"] = fixingStartStep

                if fixingEndStep is not None:
                    entry["parameters"]["endStep"] = fixingEndStep

                sim["modelExtensions"].append(copy.deepcopy(entry))

            ##Add surface
            if self.addSurface:

                if self.maxForce is not None:
                    sim["modelExtensions"].append({"type":"surfaceMaxForce","parameters":{"epsilon":epsilonSurface_smp,
                                                                                          "surfacePosition":surfacePosition_smp,
                                                                                          "maxForce":self.maxForce,
                                                                                          "selection":{"models":copy.deepcopy(sampleSelection)}}})
                else:
                    sim["modelExtensions"].append({"type":"surface","parameters":{"epsilon":epsilonSurface_smp,
                                                                                  "surfacePosition":surfacePosition_smp,
                                                                                  "selection":{"models":copy.deepcopy(sampleSelection)}}})

                if self.absorptionHeight is not None:
                    sim["modelExtensions"].append({"type":"absortionSurface","parameters":{"heightThreshold":absorptionHeight_smp,
                                                                                           "K":absorptionK_smp,
                                                                                           "startStep":thermSteps_smp}})

            #Add maxForce
            if self.maxForce is not None:
                sim["simulationSteps"].append({"type":"AFMMaxForce","parameters":{"maxForce":self.maxForce,
                                                                                  "intervalStep":self.maxForceIntervalStep}})

            #Add measures

            #Add backup
            if self.backupIntervalStep is not None:
                sim["system"].append({"type":"backup","parameters":{"backupIntervalStep":self.backupIntervalStep}})

            #Add info
            if self.infoIntervalStep is not None:
                sim["simulationSteps"].append({"type":"info","parameters":{"intervalStep":self.infoIntervalStep}})

            #Add save state
            if self.saveState:
                sim["simulationSteps"].append({"type":"saveState","parameters":{"intervalStep":self.saveStateIntervalStep,
                                                                                "outputFilePath":self.saveStateOutputFilePath,
                                                                                "outputFormat":self.saveStateOutputFormat}})

            #Add AFM measurement
            if self.afmMeasurementIntervalStep is not None:
                sim["simulationSteps"].append({"type":"afmMeasurement","parameters":{"intervalStep":self.afmMeasurementIntervalStep,
                                                                                     "outputFilePath":self.afmMeasurementOutputFilePath,
                                                                                     "startStep":thermSteps_smp}})

            simulationPool.append(copy.deepcopy(sim))

        self.loadSimulationPool(copy.deepcopy(simulationPool))

    def setUpSimulation(self, sessionName):
        super().setUpSimulation(sessionName)

class AnalysisAFM:

    def __init__(self,
                 VLMPsessionFilePath,
                 outputUnits,
                 maxForce = None,
                 plotTime = False):

        self.logger = logging.getLogger("VLMP")

        ########################################################

        self.VLMPsessionFilePath = VLMPsessionFilePath

        with open(VLMPsessionFilePath,"r") as f:
            self.VLMPsession = json.load(f)

        self.outputUnits = outputUnits
        self.maxForce    = maxForce
        self.plotTime    = plotTime

    def run(self):

        measureKey = ["simulationSteps","afmMeasurement","outputFilePath"]

        for sim in self.VLMPsession["simulations"]:
            name,_,results,info = sim

            outputFilePath = None
            if measureKey[0] in info:
                for entry in info[measureKey[0]]:
                    if entry["type"] == measureKey[1]:
                        outputFilePath = entry["parameters"][measureKey[2]]

            if outputFilePath is None:
                self.logger.warning("[AnalysisAFM] No AFM measurement found for simulation {}".format(name))
                continue

            indentationFilePath = "/".join(self.VLMPsessionFilePath.split("/")[:-1]) + "/results/" + name + "/afm.dat"
            #Check if indentation file exists
            if not os.path.isfile(indentationFilePath):
                self.logger.warning("[AnalysisAFM] Indentation file {} not found for simulation {}".format(indentationFilePath,name))
                continue
            indentationData = np.loadtxt(indentationFilePath,skiprows=1)

            T    = indentationData[:,0]
            X    = indentationData[:,1]
            F    = indentationData[:,2]

            # Input units
            inputUnits = info["units"][0]["type"]

            if inputUnits == "KcalMol_A" and self.outputUnits == "nN_nm":
                #From

                Xunits = "nm"
                Funits = "nN"

                X = X*0.1
                F = unitsUtils.KcalMol_A_force2nanonewton(F)

            f,ax = plt.subplots()
            ax.plot(X,F)
            ax.set_xlabel(f"Indentation ({Xunits})")
            ax.set_ylabel(f"Force ({Funits})")
            ax.set_title(f"Indentation for {name}")

            if self.plotTime:
                ax2 = ax.twinx()
                ax2.plot(X,T,"--",color="red")
            f.show()

            if self.maxForce is not None:
                ax.set_ylim([np.min(F),self.maxForce])

            indentationProcessedFilePath = "/".join(self.VLMPsessionFilePath.split("/")[:-1]) + "/results/" + name + "/afm_processed.dat"
            with open(indentationProcessedFilePath,"w") as f:
                f.write("# {} {}\n".format(Xunits,Funits))
                for i in range(len(X)):
                    f.write("{} {}\n".format(X[i],F[i]))

        #Wait for plots to be closed
        input("Press enter to continue...")



import sys, os

import logging

from . import simulationStepBase

class thermodynamicIntegration(simulationStepBase):
    """
    Component name: thermodynamicIntegration
    Component type: simulationStep

    Author: Pablo Ibáñez-Freire
    Date: 06/11/2023

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"outputFilePath",
                                                "stepLambda","lambdaIntervalLength",
                                                "initLambda","finalLambda"},
                         requiredParameters  = {"outputFilePath",
                                                "stepLambda","lambdaIntervalLength"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        parameters = {}

        parameters["outputFilePath"] = params["outputFilePath"]

        parameters["stepLambda"]           = params["stepLambda"]
        parameters["lambdaIntervalLength"] = params["lambdaIntervalLength"]

        parameters["initLambda"]  = params.get("initLambda",1.0)
        parameters["finalLambda"] = params.get("finalLambda",0.0)

        simulationStep = {
            name:{
              "type":["ThermodynamicMeasure","ThermodynamicIntegration"],
              "parameters":{**parameters}
            }
        }

        ############################################################

        self.setSimulationStep(simulationStep)






import sys, os

import logging

from . import modelOperationBase

import numpy as np

class setParticlePositions(modelOperationBase):

    """
    Component name: setParticlePositions
    Component type: modelOperation

    Author: Pablo Ibáñez-Freire
    Date: 31/10/2023

    Set the position of a set of particles to a given a list of ids and a position.

    :param positions: Position to set the particles to.
    :type position: list of floats

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"positions","ids"},
                         requiredParameters  = {"positions","ids"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        positions = params["positions"]
        ids       = params["ids"]

        self.setIdsState(ids,"position",positions)



import sys,os

import logging

import itertools

import psutil
import signal

import multiprocessing
import subprocess

import time
import datetime

def associateCPUtoGPU(gpuIDList):
    cpuName = multiprocessing.current_process().name
    try:
        cpuID = int(cpuName[cpuName.find('-') + 1:]) - 1
    except:
        cpuID = 0
    gpuID = gpuIDList[cpuID % len(gpuIDList)]
    return gpuID

def runSimulation(simulationsInfo, setInfo, gpuIDList):

    logger = logging.getLogger("VLMP")

    name, folder, options, components = setInfo

    gpuId = associateCPUtoGPU(gpuIDList)
    cudaFlag = 'CUDA_VISIBLE_DEVICES={}'.format(gpuId)

    fout = open(folder+"/stdout.log","w")
    ferr = open(folder+"/stderr.log","w")

    sst = time.time()

    sim = " ".join([cudaFlag, " ".join(["UAMMDlauncher",options])])
    simReturn = subprocess.run(sim, stdout=fout, stderr=ferr, shell=True, cwd=folder)

    fout.close()
    ferr.close()

    if simReturn.returncode == 0:
        logger.info("Simulation {} finished. Total time: {}".format(folder,
                                                                    str(datetime.timedelta(seconds=(time.time() - sst)))))
    else:
        logger.info("Simulation {} finished with errors. Error code: {}. Total time: {}".format(folder,
                                                                                                simReturn.returncode,
                                                                                                str(datetime.timedelta(seconds=(time.time() - sst)))))
    return simReturn

def runSimulationSets(logger, simulationsInfo, simulationSets, gpuIDList):

    with multiprocessing.Pool(processes=len(gpuIDList)) as pool:
        simSetGPU = tuple(zip(itertools.repeat(simulationsInfo.copy()),simulationSets,itertools.repeat(gpuIDList)))
        out = list(pool.starmap(runSimulation,simSetGPU))

    return out

def localLauncher(simulationSetsInfo,gpuIDList):

    simulationName  = simulationSetsInfo["name"]
    simulationsInfo = simulationSetsInfo["simulations"]
    simulationSets  = simulationSetsInfo["simulationSets"]

    logger = logging.getLogger("VLMP")

    logger.info("Start local ...")
    logger.info("Simulation name: {}".format(simulationName))
    logger.info("pid: {}".format(os.getpid()))

    def signalHandler(sig,frame):
        logger.info("Signal {} detected. Exiting ...".format(sig))

        mainProcess = psutil.Process(os.getpid())
        for child in mainProcess.children(recursive=True):
            child.kill()

        sys.exit(0)

    signal.signal(signal.SIGINT,signalHandler)
    signal.signal(signal.SIGTERM,signalHandler)

    st = time.time()
    out = runSimulationSets(logger,simulationsInfo,simulationSets,gpuIDList)
    logger.info("Simulation sets finished. Total time: {}".format(str(datetime.timedelta(seconds=(time.time() - st)))))

    for i in out:
        if(i.returncode!=0):
            logger.error("Something went wrong for simulation: {}".format(i))

    logger.info("End")

def liquidLauncher(simulationSetsInfo,nodeGPUList,postScript):

    nodeGPUList = itertools.cycle(nodeGPUList)

    user = os.environ.get("USER")

    simulationName  = simulationSetsInfo["name"]
    simulationsInfo = simulationSetsInfo["simulations"]
    simulationSets  = simulationSetsInfo["simulationSets"]

    logger = logging.getLogger("VLMP")

    logger.info("Starting liquid ...")
    logger.info("Simulation name: {}".format(simulationName))

    for jobIndex,[simSetName,simSetFolder,simSetOptions,simSetComponents] in enumerate(simulationSets):
        nodeId = next(nodeGPUList)
        jobName = f"{simulationName}_{simSetName}"

        logger.info(f'Launching simulation ...\n\
                    Job name: \"{jobName}\"\n\
                    Simulation folder: \"{simSetFolder}\"\n\
                    Binary: \"UAMMDlauncher\"\n\
                    Simulation options: \"{simSetOptions}\"\n\
                    Node: {nodeId}')

        ############################################################

        cwd = os.getcwd()
        os.chdir(simSetFolder)

        with open("./.job","w") as f:
            batch = ("#!/bin/bash\n"
                     "#$ -S /bin/bash\n"
                     f"#$ -N {jobName}\n"
                     "#$ -cwd\n"
                     "#$ -o stdout.log\n"
                     "#$ -e stderr.log\n"
                     "#$ -l gpu=1\n"
                     "#$ -V\n"
                     f"cd /scratch/{user}/{jobName}-$JOB_ID\n"
                     "module load gcc/8.4 cuda/10.2\n"
                     f"UAMMDlauncher {simSetOptions}\n"
                     f"{postScript}\n")

            f.write(batch)

        subprocess.run(["bsub","-N",jobName,
                        "-q","gpu.q",
                        "-l",f"hostname={nodeId}",
                        "-o",f"{os.getcwd()}/stdout.log",
                        "-e",f"{os.getcwd()}/stderr.log",
                        ".job"])
        time.sleep(1.0)
        os.chdir(cwd)

        ############################################################

    logger.info("All simulation sets job have been submitted")

def slurmLauncher(simulationSetsInfo,nodeList,filling,partitionList,modules,postScript):

    if modules[0] != "":
        modules = " ".join(modules)
        modules = f"module purge\n module load {modules}\n"

    lenNodeList      = len(nodeList)
    lenFilling       = len(filling)
    lenPartitionList = len(partitionList)

    if(lenPartitionList == 1 and lenNodeList >1):
        part = partitionList[0]
        partitionList    = [part]*lenNodeList
        lenPartitionList = len(partitionList)

    if(lenNodeList == 0):
        nodeList = [None]*lenPartitionList
    else:
        if(lenNodeList!=lenPartitionList):
            self.logger.error("The number of nodes and the number of partitions must be the same")
            sys.exit(1)

        if(lenNodeList!=lenFilling):
            self.logger.error("The number of nodes and the number of filling entries must be the same")
            sys.exit(1)

        nodeListTmp      = []
        partitionListTmp = []
        for i,f in enumerate(filling):
            for _ in range(f):
                nodeListTmp.append(nodeList[i])
                partitionListTmp.append(partitionList[i])

        nodeList      = nodeListTmp.copy()
        partitionList = partitionListTmp.copy()

    nodePartitionList = list(zip(nodeList,partitionList))
    nodePartitionList = itertools.cycle(nodePartitionList)

    simulationName  = simulationSetsInfo["name"]
    simulationsInfo = simulationSetsInfo["simulations"]
    simulationSets  = simulationSetsInfo["simulationSets"]

    logger = logging.getLogger("VLMP")

    logger.info("Starting slurm ...")
    logger.info("Simulation name: {}".format(simulationName))

    jobIds = []

    for jobIndex,[simSetName,simSetFolder,simSetOptions,simSetComponents] in enumerate(simulationSets):
        node,partition = next(nodePartitionList)

        jobName = f"{simulationName}_{simSetName}"

        logger.info(f'Launching simulation ...\n\
                    Job name: \"{jobName}\"\n\
                    Simulation folder: \"{simSetFolder}\"\n\
                    Binary: \"UAMMDlauncher\"\n\
                    Simulation options: \"{simSetOptions}\"\n\
                    Node: \"{node}\"\n\
                    Partition: \"{partition}\"')

        print("Launching job: {}".format(jobName))

        ############################################################

        cwd = os.getcwd()
        os.chdir(simSetFolder)

        if node == None:
            nodeSBATCH = ""
        else:
            nodeSBATCH = f"#SBATCH --nodelist={node}"

        with open("./.job","w") as f:
            batch = ("#!/bin/bash\n"
                     f"#SBATCH --job-name={jobName}\n"
                     f"#SBATCH --partition={partition}\n"
                      "#SBATCH --nodes=1\n"
                      "#SBATCH --ntasks-per-node=1\n"
                      "#SBATCH --cpus-per-task=1\n"
                      "#SBATCH --gres=gpu:1\n"
                      "#SBATCH --output=stdout.log\n"
                      "#SBATCH --error=stderr.log\n"
                     f"{nodeSBATCH}\n"
                     f"{modules}\n"
                     f"UAMMDlauncher {simSetOptions}\n"
                     f"{postScript}\n")

            f.write(batch)

        result = subprocess.run(["sbatch", ".job"], capture_output=True, text=True)

        # The output is in result.stdout
        if result.returncode == 0:  # Checking if the sbatch command was successful
            output = result.stdout.strip()
            # The job ID is typically in the format "Submitted batch job <job_id>"
            if "Submitted batch job" in output:
                job_id = output.split()[-1]  # Extracting the job ID from the output
                logger.info(f"Job submitted successfully. Job ID: {job_id}")
                jobIds.append(job_id)
            else:
                logger.error("Failed to submit job")
                logger.error(output)
                sys.exit(1)
        else:
            logger.error("Failed to submit job")
            logger.error(result.stderr)
            sys.exit(1)

        time.sleep(1.0)

        os.chdir(cwd)

        ############################################################
    return jobIds

def rebuildResults(simulationSetsInfo):

    logger = logging.getLogger("VLMP")

    #Check if results folder exists
    if os.path.exists("results"):
        logger.error("The results folder already exists")
        sys.exit(1)
    else:
        os.mkdir("results")

    simulationName  = simulationSetsInfo["name"]
    simulationsInfo = simulationSetsInfo["simulations"]
    simulationSets  = simulationSetsInfo["simulationSets"]

    for simName,simSet,simResult,_ in simulationsInfo:
        #Create a symbolic link to the simulation results
        os.symlink(f"../{simSet}",f"./{simResult}")




from VLMP.components.models import modelBase

import numpy as np
#from scipy.spatial import cKDTree

from icosphere import icosphere

class MAGNETICNP(modelBase):
    """
    Component name: MAGNETICNP
    Component type: model

    Author: P. Palacios-Alonso
    Date: 16/10/2023

    Model of magnetic nanoparticles
    """

    def __generateLogNormalDistribution(self, mean, std, N):
        distribution = np.full(N, mean)

        if (std>0.0):
            mean2 = mean*mean;
            std2  = std*std;

            sigma = (np.log(1+std2/(mean2)))**0.5;
            mu    = np.log(mean2/(mean2+std2)**0.5);

            distribution = np.random.lognormal(mean  = mu,
                                               sigma = sigma,
                                               size  = N)
        return distribution


    def __axisToQuaternion(self, axis):
        # Normalize the axis
        axis = axis / np.sqrt(np.dot(axis, axis))
        #When axis is parallel to u_z, v is zero and the have to be computed in a different way.
        if axis[2] == 1.0:
            return np.array([1.0, 0.0, 0.0, 0.0])
        if axis[2] == -1.0:
            return np.array([0.0, 1.0, 0.0, 0.0])

        phi = np.arccos(axis[2])
        # u_z x axis (cross product)

        v = np.array([-axis[1], axis[0], 0.0])
        v = v / np.sqrt(1 - axis[2]**2)

        # Compute sin and cos of phi/2
        sphi_2 = np.sin(phi * 0.5)
        cphi_2 = np.cos(phi * 0.5)

        qdir = np.array([cphi_2, sphi_2 * v[0], sphi_2 * v[1], sphi_2 * v[2]])
        return qdir


    def __generateRandomQuaternions(self, N):
        x0 = np.random.rand(N)
        r1 = np.sqrt(1.0-x0);
        r2 = np.sqrt(x0);
        ang1 = 2*np.pi*np.random.rand(N)
        ang2 = 2*np.pi*np.random.rand(N)
        return np.column_stack((r2*np.cos(ang2),
                                r1*np.sin(ang1),
                                r1*np.cos(ang1),
                                r2*np.sin(ang2)))



    def __generateInitialOrientations(self, N, mode, axis = [0,0,1]):
        if mode == "aligned":
            axis       = np.array(axis)
            quaternion = self.__axisToQuaternion(axis)
            qdirs      = np.tile(quaternion, (N, 1))
        elif mode == "random":
            qdirs = self.__generateRandomQuaternions(N)
        else:
            self.logger.error(f"[MAGNETICNP] Initial orientation {mode} not valid")
            raise Exception("Initial orientation not available")

        return qdirs

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"particleName",
                                                "nParticles",
                                                "msat",
                                                "anisotropy",
                                                "anisotropyStd",
                                                "coreRadius",
                                                "coreRadiusStd",
                                                "coatingWidth",
                                                "coatingWidthStd",
                                                "initOrientation",
                                                "initAxis"},
                         requiredParameters  = {"nParticles", "msat", "coreRadius"},
                         definedSelections   = {"particleId"},
                         **params)

        ############################################################

        partName = params.get("particleName", "A")

        msat           = params["msat"]
        coreRadiusMean = params["coreRadius"]
        nParticles     = params["nParticles"]

        anisotropyMean   = params.get("anisotropy"     , None)
        anisotropyStd    = params.get("anisotropyStd"  , 0.0)
        coreRadiusStd    = params.get("coreRadiusStd"  , 0.0)
        coatingWidthMean = params.get("coatingWidth"   , 0.0)
        coatingWidthStd  = params.get("coatingWidthStd", 0.0)
        initOrientation  = params.get("initOrientation", "aligned")
        initAxis         = params.get("initAxis"       , [0,0,1])

        coreRadius   = self.__generateLogNormalDistribution(coreRadiusMean,
                                                            coreRadiusStd,
                                                            nParticles)

        coatingWidth = self.__generateLogNormalDistribution(coatingWidthMean,
                                                            coatingWidthStd,
                                                            nParticles)

        magneticMoment     = 4./3.*np.pi*coreRadius**3*msat
        hydrodynamicRadius = coreRadius + coatingWidth

        if anisotropyMean is not None:
            anisotropy = self.__generateLogNormalDistribution(anisotropyMean,
                                                              anisotropyStd,
                                                              nParticles)



        directions = self.__generateInitialOrientations(nParticles, initOrientation, initAxis)
        types = self.getTypes()
        types.addType(name = partName)

        state = {}
        if anisotropyMean is not None:
            state["labels"] = ["id", "radius", "position", "direction", "magnetization", "anisotropy"]
            state["data"]   = []
            for i in range(nParticles):
                state["data"].append([i,
                                      hydrodynamicRadius[i],
                                      [0,0,0], directions[i].tolist(),
                                      [0,0,1,magneticMoment[i]],
                                      anisotropy[i]])
        else:
            state["labels"] = ["id", "radius", "position", "direction", "magnetization"]
            state["data"] = []
            for i in range(nParticles):
                state["data"].append([i,
                                      hydrodynamicRadius[i],
                                      [0,0,0], directions[i].tolist(),
                                      [0,0,1,magneticMoment[i]]])
        structure = {}
        structure["labels"] = ["id","type"]
        structure["data"] = []
        for i in range(nParticles):
            structure["data"].append([i,partName])


        #Generate forceField

        self.setState(state)
        self.setStructure(structure)

    def processSelection(self,**params):

        sel = []
        if "particleId" in params:
            sel += params["particleId"]
        return sel



from VLMP.components.models import modelBase

import os

import json

import numpy as np
import icosphere

class ICOSPHERE(modelBase):
    """
    Component name: ICOSPHERE
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 18/06/2023

    Icosphere model
    """

    def __loadData(self,name):

        try:

            with open(self.modelData,"r") as f:
                data = json.load(f)
        except:
            self.logger.warning("[ICOSPHERE] Error opening model data from {} (loading)".format(self.modelData))
            return None

        try:
            d = data[str(self.resolution)][name].copy()
        except:
            self.logger.warning("[ICOSPHERE] Error loading {} from model data".format(name))
            return None

        return d

    def __saveToData(self,d,name):

        try:
            with open(self.modelData,"r") as f:
                data = json.load(f)
        except:
            self.logger.warning("[ICOSPHERE] Error opening model data from {} (saving)".format(self.modelData))
            return

        try:
            if str(self.resolution) not in data:
                data[str(self.resolution)] = {}
            data[str(self.resolution)][name] = d.copy()

            with open(self.modelData,"w") as f:
                json.dump(data,f)
        except:
            self.logger.warning("[ICOSPHERE] Error saving model data to {}".format(self.modelData))
            return

        return

    def __generateBonds(self,faces):

        bonds = self.__loadData("bonds")

        if bonds is None:
            bonds = set()
            for face in faces:
                bonds.add(tuple(sorted([int(face[0]),int(face[1])])))
                bonds.add(tuple(sorted([int(face[1]),int(face[2])])))
                bonds.add(tuple(sorted([int(face[2]),int(face[0])])))
            bonds = list(bonds)

            self.__saveToData(bonds,"bonds")

        return bonds.copy()

    def __generateDihedrals(self,faces):

        dihedrals = self.__loadData("dihedrals")

        if dihedrals is None:

            dihedrals = []

            # List all pairs of faces that share an edge
            facesPairs = set()
            for i in range(len(faces)):
                for j in range(i+1,len(faces)):
                    if len(set(faces[i]).intersection(set(faces[j]))) == 2:
                        facesPairs.add((i,j))

            for facePair in facesPairs:
                face1 = faces[facePair[0]]
                face2 = faces[facePair[1]]

                common = set(face1).intersection(set(face2))

                if len(common) == 2:
                    face1Id = list(set(face1).difference(common))[0]
                    face2Id = list(set(face2).difference(common))[0]

                    i = int(face1Id)
                    j = int(common.pop())
                    k = int(common.pop())
                    l = int(face2Id)

                    dihedrals.append([i,j,k,l])
                else:
                    self.log("Faces do not share an edge, faces: {} {}".format(face1,face2))
                    raise Exception("Faces do not share an edge")

            self.__saveToData(dihedrals,"dihedrals")

        return dihedrals.copy()

    def __generateExclusions(self,bonds):

        exclusions = self.__loadData("exclusions")

        if exclusions is None:

            exclusionsTmp = {}

            for bnd in bonds:
                i,j = bnd
                exclusionsTmp[i] = []
                exclusionsTmp[j] = []
            for bnd in bonds:
                i,j = bnd
                exclusionsTmp[i].append(j)
                exclusionsTmp[j].append(i)

            exclusionsTmp = {i:sorted(list(set(data))) for i,data in exclusionsTmp.items()}

            exclusions = []
            for i,excl in exclusionsTmp.items():
                exclusions.append([i,excl.copy()])

            #Sort according i
            exclusions = sorted(exclusions,key=lambda x: x[0])

            self.__saveToData(exclusions,"exclusions")

        return exclusions.copy()

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"particleName",
                                                "particleMass","particleRadius","particleCharge",
                                                "position",
                                                "resolution",
                                                "radius",
                                                "Kb",
                                                "Kd",
                                                "steric"},
                         requiredParameters  = {"particleName"},
                         definedSelections   = {"particleId"},
                         **params)

        ############################################################

        self.modelData = "./data/ICOSPHERE.json"
        self.modelData = os.path.join(os.path.dirname(os.path.realpath(__file__)),self.modelData)

        ############################################################

        particleName = params["particleName"]

        particleMass   = params.get("particleMass",1.0)
        particleRadius = params.get("particleRadius",1.0)
        particleCharge = params.get("particleCharge",0.0)

        types = self.getTypes()
        types.addType(name = particleName,
                      mass = particleMass,
                      radius = particleRadius,
                      charge = particleCharge)

        position = params.get("position",np.array([0.0,0.0,0.0]))

        self.resolution = params.get("resolution",1)
        radius          = params.get("radius",1.0)

        Kb = params.get("Kb",1.0)
        Kd = params.get("Kd",0.0)

        steric = params.get("steric",False)

        ############################################################

        vertices,faces = icosphere.icosphere(nu = self.resolution)

        #Update the vertices
        for i in range(len(vertices)):
            vertices[i] = radius*vertices[i]+position

        state = {}
        state["labels"] = ["id","position"]

        state["data"] = []
        for i,vertex in enumerate(vertices):
            x,y,z = vertex
            state["data"].append([i,[x,y,z]])

        structure = {}
        structure["labels"] = ["id","type"]

        structure["data"] = []
        for i,face in enumerate(vertices):
            structure["data"].append([i,particleName])

        forceField = {}

        forceField["bonds"] = {}
        forceField["bonds"]["type"]       = ["Bond2","Harmonic"]
        forceField["bonds"]["labels"]     = ["id_i","id_j","r0","K"]
        forceField["bonds"]["parameters"] = {}
        forceField["bonds"]["data"]       = []

        bonds = self.__generateBonds(faces)

        for bond in bonds:
            dst = np.linalg.norm(vertices[bond[0]]-vertices[bond[1]])
            forceField["bonds"]["data"].append([bond[0],bond[1],dst,Kb])

        if Kd > 0.0:

            forceField["dihedrals"] = {}
            forceField["dihedrals"]["type"]       = ["Bond4","Dihedral"]
            forceField["dihedrals"]["labels"]     = ["id_i","id_j","id_k","id_l","phi0","K","n"]
            forceField["dihedrals"]["parameters"] = {}
            forceField["dihedrals"]["data"]       = []

            dihedrals = self.__generateDihedrals(faces)

            for dihedral in dihedrals:
                i,j,k,l = dihedral
                forceField["dihedrals"]["data"].append([i,j,k,l,0.0,Kd,1])

        if steric:

            exclusions = self.__generateExclusions(bonds)

            forceField["verletList"] = {}
            forceField["verletList"]["type"]       = ["VerletConditionalListSet", "nonExclIntra_nonExclInter"]
            forceField["verletList"]["parameters"] = {"cutOffVerletFactor":1.2}
            forceField["verletList"]["labels"]     = ["id","id_list"]
            forceField["verletList"]["data"]       = []

            for i,excl in exclusions:
                forceField["verletList"]["data"].append([i,excl])

            forceField["steric"] = {}
            forceField["steric"]["type"] = ["NonBonded", "WCAType2"]
            forceField["steric"]["parameters"] = {"cutOffFactor": 2.5,"condition":"intra"}
            forceField["steric"]["labels"] = ["name_i", "name_j", "epsilon", "sigma"]
            forceField["steric"]["data"]   = []

            forceField["steric"]["data"].append([particleName,particleName,1.0,pow(2.0,1.0/6.0)*particleRadius])

        ############################################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)


    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += params["particleId"]

        return sel




import sys, os

import copy

import logging

from . import integratorBase

class EulerMaruyama(integratorBase):

    """
    Component name: EulerMaruyama
    Component type: integrator

    Author: Pablo Ibáñez-Freire
    Date: 04/04/2023

    Simple Euler-Maruyama integrator for brownian dynamics.

    :param integrationSteps: Number of integration steps.
    :type integrationSteps: int
    :param timeStep: Time step of the integrator.
    :type timeStep: float
    :param viscosity: Viscosity of the fluid.
    :type viscosity: float


    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"integrationSteps","timeStep","viscosity"},
                         requiredParameters  = {"integrationSteps","timeStep","viscosity"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        integrationSteps = params.get("integrationSteps")

        parameters = {}

        parameters["timeStep"]         = params.get("timeStep")
        parameters["viscosity"]        = params.get("viscosity")

        integrator = {
            "type" : ["Brownian", "EulerMaruyama"],
            "parameters" : copy.deepcopy(parameters)
        }

        ############################################################

        self.setIntegrationSteps(integrationSteps)
        self.setIntegrator(integrator)



import logging

def getLabelIndex(l,labels):
    logger = logging.getLogger("VLMP")

    if l in labels:
        return labels.index(l)
    else:
        logger.error("Label %s not found in labels list" % l)
        raise Exception("Label not found")

def getSubParameters(prmName,parameters):

    logger = logging.getLogger("VLMP")

    param = parameters.get(prmName,{})
    if not isinstance(param,dict):
        #Check if the parameter is a string
        if isinstance(param,str):
            param = {param:{}}
        else:
            logger.error(f"Parameter {prmName} is not a dictionary nor a string")
            raise Exception("Parameter is not a dictionary")

    parameterName   = None
    subParameters   = {}

    if len(param) == 0:
        return parameterName, subParameters
    if len(param) > 1:
        logger.error("Only one subparameter can be specified")
        raise Exception("Only one subparameter can be specified")

    parameterName = list(param.keys())[0]
    subParameters = param[parameterName]

    return parameterName, subParameters



import sys, os

import logging

from . import simulationStepBase

class vqcmMeasurement(simulationStepBase):
    """
    Component name: vqcmMeasurement
    Component type: simulationStep

    Author: Pablo Palacios-Alonso and Pablo Ibáñez-Freire
    Date: 2/11/2023

    """


    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"outputFilePath",
                                                "kernel",
                                                "tolerance",
                                                "f0",
                                                "overtone",
                                                "hydrodynamicRadius",
                                                "viscosity",
                                                "vwall",
                                                "fluidDensity",
                                                "maxNIterations",
                                                "toleranceConvergence",
                                                "maxNIterations",
                                                "memory",
                                                "dampingAnderson",
                                                "dampingPiccard",
                                                "notAcceleratedInterval",
                                                "h",
                                                "printSteps"},
                         requiredParameters  = {"outputFilePath",
                                                "f0","overtone",
                                                "hydrodynamicRadius",
                                                "viscosity",
                                                "vwall",
                                                "fluidDensity"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        parameters = {}

        parameters["outputFilePath"] = params["outputFilePath"]

        parameters["kernel"]    = params.get("kernel", "Peskin3p")
        if "tolerance" in params and parameters["kernel"] == "Peskin3p":
            self.logger.error("[vqcmMeasurement] tolerance parameter is not available for Peskin3p kernel")
        if parameters["kernel"] != "Peskin3p":
            parameters["tolerance"] = params.get("tolerance", 1e-5)

        parameters["f0"]                 = params["f0"]
        parameters["overtone"]           = params["overtone"]
        parameters["hydrodynamicRadius"] = params["hydrodynamicRadius"]
        parameters["viscosity"]          = params["viscosity"]
        parameters["vwall"]              = params["vwall"]
        parameters["fluidDensity"]       = params["fluidDensity"]

        parameters["maxNIterations"]         = params.get("maxNIterations", 10000)
        parameters["toleranceConvergence"]   = params.get("toleranceConvergence", 1e-4)
        parameters["memory"]                 = params.get("memory", 5)
        parameters["dampingAnderson"]        = params.get("dampingAnderson", 1e-5)
        parameters["dampingPiccard"]         = params.get("dampingPiccard", 1e-5)
        parameters["notAcceleratedInterval"] = params.get("notAcceleratedInterval", 2)
        parameters["h"]                      = params.get("h", 0.0) # 0.0 means it is computed by UAMMD
        
        if "printSteps" in params:
            parameters["printSteps"] = params["printSteps"]

        simulationStep = {
            name:{
              "type":["VQCMMeasure","VQCMMeasure"],
              "parameters":{**parameters}
            }
        }

        ############################################################

        self.setSimulationStep(simulationStep)






import os
import copy

import logging

################ MODEL INTERFACE ################

from pyUAMMD import simulation

from .. import idsHandler

from ...utils.selections import getSelections

class modelExtensionBase(idsHandler):

    def __init__(self,
                 _type:str,_name:str,
                 units,types,ensemble,
                 models,
                 availableParameters:set,
                 requiredParameters:set,
                 availableSelections:set,
                 requiredSelections:set,
                 **params):
        super().__init__(models)

        self.logger = logging.getLogger("VLMP")

        self._type = _type
        self._name = _name

        self._units  = units
        self._types  = types
        self._ensemble = ensemble

        self._models = models

        self.logger.debug(f"[ModelExtension] ({self._type}) Extending models: "+
                          " ".join([m.getName() for m in self._models])+
                          ". For model extension: "+self._name)

        self.availableParameters =  availableParameters.copy()
        self.availableParameters.update({"startStep","endStep"})
        self.availableSelections =  availableSelections.copy()

        self.requiredParameters  =  requiredParameters.copy()
        self.requiredSelections  =  requiredSelections.copy()

        # Check all required parameters are available parameters
        if not self.requiredParameters.issubset(self.availableParameters):
            notAvailable = self.requiredParameters.difference(self.availableParameters)
            self.logger.error(f"[ModelExtension] ({self._type}) Some required parameters ({notAvailable}) are not available parameters for model extension {self._name}")
            raise Exception(f"Required paramaters are not available parameters")

        # Check all required selections are available selections
        if not self.requiredSelections.issubset(self.availableSelections):
            notAvailable = self.requiredSelections.difference(self.availableSelections)
            self.logger.error(f"[ModelExtension] ({self._type}) Some required selections ({notAvailable}) are not available selections for model extension {self._name}")
            raise Exception(f"Required selections are not available selections")

        # Check if all parameters and selectors given by params are available
        for par in params:
            if par not in self.availableParameters and par not in self.availableSelections:
                self.logger.error(f"[ModelExtension] ({self._type}) Parameter or selection {par} not available for model extension {self._name}")
                raise Exception(f"Parameter not available")

        # Check if all required parameters are given
        for par in self.requiredParameters:
            if par not in params:
                self.logger.error(f"[ModelExtension] ({self._type}) Required parameter {par} not given for model extension {self._name}")
                raise Exception(f"Required parameter not given")

        # Check if all required selections are given
        for sel in self.requiredSelections:
            if sel not in params:
                self.logger.error(f"[ModelExtension] ({self._type}) Required selection {sel} not given for model extension {self._name}")
                raise Exception(f"Required selection not given")

        self.logger.info(f"[ModelExtension] ({self._type}) Using model extension {self._name}")

        ########################################################

        self._startStep = params.get("startStep",None)
        self._endStep   = params.get("endStep",None)

        ########################################################

        #Process selections
        selections = [sel for sel in params if sel in self.availableSelections]
        self._selection = getSelections(self._models,
                                        selections,
                                        **params)

        ########################################################

        self._extension = None
        self._group = None

    ########################################################

    def getName(self):
        return self._name

    def getType(self):
        return self._type

    ########################################################

    def getUnits(self):
        return self._units

    def getTypes(self):
        return self._types

    def getEnsemble(self):
        return self._ensemble

    def getModels(self):
        return self._models

    ########################################################

    def setExtension(self,extension):
        self._extension = extension

    def getExtension(self):
        if self._extension is None:
            self.logger.error(f"[ModelExtension] ({self._type}) Extension not initialized")
            raise Exception(f"Extension not initialized")

        if self._startStep is not None:
            self._extension[self.getName()]["parameters"]["startStep"] = self._startStep
        if self._endStep is not None:
            self._extension[self.getName()]["parameters"]["endStep"]   = self._endStep

        return self._extension

    def setGroup(self,selSelections):

        #If selSelections is not a list, make it a list
        if not isinstance(selSelections,list):
            selSelections = [selSelections]

        #Check if all sel in selSelections are available selections
        if not set(selSelections).issubset(self.availableSelections):
            notAvailable = set(selSelections).difference(self.availableSelections)
            self.logger.error(f"[ModelExtension] ({self._type}) Some selections ({notAvailable}),"
                              f" requested to set a group, are not available selections for modelExtension {self._name}")
            raise Exception(f"Selections not available")

        selections = [sel for sel in self._selection.keys() if sel in selSelections]

        ids = []
        for sel in selections:
            ids.extend(self.getSelection(sel))

        #If ids is empty, then no particles were selected. Do nothing
        if len(ids) == 0:
            return

        self._group = {"type":["Groups","GroupsList"],
                       "parameters":{},
                       "labels":["name","type","selection"],
                       "data":[[self.getName(),"Ids",list(set(ids)).copy()]]}

    ########################################################

    def getSelection(self,selectionName):
        return self._selection[selectionName]

    ########################################################

    def getIdsProperty(self,ids,propertyName):
        return self._getIdsProperty(ids,propertyName)

    def getIdsState(self,ids,stateName):
        return self._getIdsState(ids,stateName)

    def getIdsStructure(self,ids,structName):
        return self._getIdsStructure(ids,structName)

    ########################################################

    def getSimulation(self,DEBUG_MODE = False):

        sim = {}

        sim["topology"] = {}
        sim["topology"]["forceField"] = self.getExtension()

        if self._group is not None:
            groupName = "group_"+self.getName()
            for ext in sim["topology"]["forceField"]:
                sim["topology"]["forceField"][ext]["parameters"]["group"] = groupName
            sim["topology"]["forceField"][groupName] = self._group

        return simulation(copy.deepcopy(sim),DEBUG_MODE)

############### IMPORT ALL MODEL EXTENSIONS ###############

import glob

currentPath = os.path.dirname(os.path.abspath(__file__))
extensions = [os.path.basename(x) for x in glob.glob(currentPath+"/*.py")]
extensions = [x.split(".")[0] for x in extensions if "__" not in x]

for e in extensions:
    try:
        exec(f"from .{e} import *")
    except Exception as e:
        logging.getLogger("VLMP").error(e)
        logging.getLogger("VLMP").error(f"[ModelExtension] Error importing model extension type component {e}")



import os
import copy

import logging

################ SYSTEM INTERFACE ################

from pyUAMMD import simulation

class systemBase:

    def __init__(self,
                 _type:str,_name:str,
                 availableParameters:set,
                 requiredParameters:set,
                 **params):

        self.logger = logging.getLogger("VLMP")

        self._type = _type
        self._name = _name

        self.availableParameters = availableParameters.copy()
        self.requiredParameters  = requiredParameters.copy()

        # Check all required parameters are available parameters
        if not self.requiredParameters.issubset(self.availableParameters):
            notAvailable = self.requiredParameters.difference(self.availableParameters)
            self.logger.error(f"[System] ({self._type}) Some required parameters ({notAvailable}) are not available parameters for system {self._name}")
            raise Exception(f"Required paramaters are not available parameters")

        # Check if all parameters given by params are available
        for par in params:
            if par not in self.availableParameters:
                self.logger.error(f"[System] ({self._type}) Parameter {par} not available for system {self._name}")
                raise Exception(f"Parameter not available")

        # Check if all required parameters are given
        for par in self.requiredParameters:
            if par not in params:
                self.logger.error(f"[System] ({self._type}) Required parameter {par} not given for system {self._name}")
                raise Exception(f"Required parameter not given")

        self.logger.info(f"[System] ({self._type}) Using system {self._name}")

        ########################################################

        self._system = None

    ########################################################

    def getName(self):
        return self._name

    def getType(self):
        return self._type

    ########################################################

    def setSystem(self,system):
        self._system = system

    def getSystem(self):
        if self._system is None:
            self.logger.error(f"[System] ({self._type}) System {self._name} not initialized")
            raise Exception(f"System not initialized")
        return self._system

    ########################################################

    def getSimulation(self,DEBUG_MODE = False):
        return simulation({"system":copy.deepcopy(self.getSystem())},DEBUG_MODE)

############### IMPORT ALL SYSTEMS ###############

import glob

currentPath = os.path.dirname(os.path.abspath(__file__))
#Get file name, without full path
systems = [os.path.basename(x) for x in glob.glob(currentPath+"/*.py")]
systems = [x.split(".")[0] for x in systems if "__" not in x]

for s in systems:
    try:
        exec(f"from .{s} import *")
    except Exception as e:
        logging.getLogger("VLMP").error(e)
        logging.getLogger("VLMP").error(f"[System] Error importing system type component \"{s}\"")




import os
import copy

import logging

################ MODEL INTERFACE ################

import abc
from pyUAMMD import simulation

from ...utils.input import getLabelIndex

class modelBase(metaclass=abc.ABCMeta):

    def __init__(self,
                 _type:str,_name:str,
                 units,types,ensemble,
                 availableParameters:set,
                 requiredParameters:set,
                 definedSelections:set,
                 **params):

        self.logger = logging.getLogger("VLMP")

        self._type = _type
        self._name = _name

        self._units    = units
        self._types    = types
        self._ensemble = ensemble

        self.availableParameters = availableParameters.copy()
        self.requiredParameters  = requiredParameters.copy()
        self.definedSelections   = definedSelections.copy()

        # Check all required parameters are available parameters
        if not self.requiredParameters.issubset(self.availableParameters):
            notAvailable = self.requiredParameters.difference(self.availableParameters)
            self.logger.error(f"[Model] ({self._type}) Some required parameters ({notAvailable}) are not available parameters for model {self._name}")
            raise Exception(f"Required paramaters are not available parameters")

        # Check if all parameters given by params are available
        for par in params:
            if par not in self.availableParameters:
                self.logger.error(f"[Model] ({self._type}) Parameter {par} not available for model {self._name}")
                raise Exception(f"Parameter not available")

        # Check if all required parameters are given
        for par in self.requiredParameters:
            if par not in params:
                self.logger.error(f"[Model] ({self._type}) Required parameter {par} not given for model {self._name}")
                raise Exception(f"Required parameter not given")

        self.logger.info(f"[Model] ({self._type}) Using model {self._name}")

        ########################################################

        self._idOffset   = None

        ########################################################

        self._state      = None
        self._structure  = None
        self._forceField = None

    ########################################################

    def getName(self):
        return self._name

    def getType(self):
        return self._type

    ########################################################

    def getUnits(self):
        return self._units

    def getTypes(self):
        return self._types

    def getEnsemble(self):
        return self._ensemble

    ########################################################

    def setState(self,state):
        self._state = state

    def setStructure(self,structure):
        self._structure = structure

    def setForceField(self,forceField):
        self._forceField = forceField

    def getState(self):
        if self._state is None:
            self.logger.error(f"[Model] ({self._type}) State not set")
            raise Exception(f"State not set")
        return self._state

    def getStructure(self):
        if self._structure is None:
            self.logger.error(f"[Model] ({self._type}) Structure not set")
            raise Exception(f"Structure not set")
        return self._structure

    def getForceField(self):
        if self._forceField is None:
            self.logger.error(f"[Model] ({self._type}) Force field not set")
            raise Exception(f"Force field not set")
        return self._forceField

    ########################################################

    def getNumberOfParticles(self):
        if self._state is None:
            return 0
        return len(self.getState()["data"])

    def getLocalIds(self):
        if self._state is None:
            return []
        ids = []
        idIndex = getLabelIndex("id",self.getState()["labels"])
        for entry in self.getState()["data"]:
            ids.append(entry[idIndex])
        return ids

    def setIdOffset(self,offset):
        self._idOffset = offset

    def getIdOffset(self):
        if self._idOffset is None:
            self.logger.error(f"[Model] ({self._type}) Id offset not set")
            raise Exception(f"Id offset not set")
        return self._idOffset

    def getGlobalIds(self):
        globalIds = []
        for localId in self.getLocalIds():
            globalIds.append(localId + self.getIdOffset())
        return globalIds

    ########################################################

    def getSimulation(self,DEBUG_MODE = False):

        # Create simulation

        sim = {}

        if self._state is not None:
            sim["state"]  = self.getState()

        sim["topology"] = {}
        if self._structure is not None:
            sim["topology"]["structure"]  = self.getStructure()
        if self._forceField is not None:
            sim["topology"]["forceField"] = self.getForceField()

        return simulation(copy.deepcopy(sim),DEBUG_MODE)

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'processSelection') and
                callable(subclass.processSelection)   and
                NotImplemented)

    @abc.abstractmethod
    def processSelection(self,**params):
        """ Return a index of the particles that are selected """
        raise NotImplementedError

    def getSelection(self,**params):
        # If no selection is given, select all particles
        if len(params) == 0:
            return self.getLocalIds()
        # Check if all selectors given by params are available
        for par in params:
            if par not in self.definedSelections:
                self.logger.error(f"[Model] ({self._type}) Selector {par} not available for model {self._name}")
                raise Exception(f"Selector not available")
        return self.processSelection(**params)

    ########################################################
    #Common selections functions

    def getParticleIdSelection(self,particleId):
        #Check if particleId is a list
        if not isinstance(particleId,list):
            particleId = [particleId]

        localIds = self.getLocalIds()
        # Check if particleId is in localIds
        for pid in particleId:
            if pid not in localIds:
                self.logger.error(f"[Model] ({self._type}) Particle id {pid} not in local ids")
                raise Exception(f"Particle id not in local ids")

        return particleId.copy()

    def getForceFieldSelection(self,forceFieldEntriesName):
        #Check if forceFieldEntriesName is a list
        if not isinstance(forceFieldEntriesName,list):
            forceFieldEntriesName = [forceFieldEntriesName]

        sel = []

        forceField = self.getForceField()
        # Check if forceFieldEntriesName is in forceField
        for ff in forceFieldEntriesName:
            if ff not in forceField:
                self.logger.error(f"[Model] ({self._type}) Force field entry {ff} not in force field")
                raise Exception(f"Force field entry not in force field")
            else:
                entryType = forceField[ff]["type"][0]

                entryLabels = forceField[ff]["labels"]
                entryData   = forceField[ff]["data"]

                if entryType == "Bond1":
                    id1Index = getLabelIndex("id_i",entryLabels)
                    for i in range(len(entryData)):
                        id1 = int(entryData[i][id1Index])
                        sel.append(id1)

                elif entryType == "Bond2":
                    id1Index = getLabelIndex("id_i",entryLabels)
                    id2Index = getLabelIndex("id_j",entryLabels)
                    for i in range(len(entryData)):
                        id1 = int(entryData[i][id1Index])
                        id2 = int(entryData[i][id2Index])
                        sel.append([id1,id2])

                elif entryType == "Bond3":
                    id1Index = getLabelIndex("id_i",entryLabels)
                    id2Index = getLabelIndex("id_j",entryLabels)
                    id3Index = getLabelIndex("id_k",entryLabels)
                    for i in range(len(entryData)):
                        id1 = int(entryData[i][id1Index])
                        id2 = int(entryData[i][id2Index])
                        id3 = int(entryData[i][id3Index])
                        sel.append([id1,id2,id3])

                elif entryType == "Bond4":
                    id1Index = getLabelIndex("id_i",entryLabels)
                    id2Index = getLabelIndex("id_j",entryLabels)
                    id3Index = getLabelIndex("id_k",entryLabels)
                    id4Index = getLabelIndex("id_l",entryLabels)
                    for i in range(len(entryData)):
                        id1 = int(entryData[i][id1Index])
                        id2 = int(entryData[i][id2Index])
                        id3 = int(entryData[i][id3Index])
                        id4 = int(entryData[i][id4Index])
                        sel.append([id1,id2,id3,id4])

                else:
                    self.logger.error(f"[Model] ({self._type}) Force field entry {ff} has not an available type."
                                       "Available types are: Bond1, Bond2, Bond3, Bond4")
                    raise Exception(f"Force field entry has not an available type")

        return sel.copy()


############### IMPORT ALL MODELS ###############

import glob

currentPath = os.path.dirname(os.path.abspath(__file__))
models = [os.path.basename(x) for x in glob.glob(currentPath+"/*.py")]
models = [x.split(".")[0] for x in models if "__" not in x]

for m in models:
    try:
        exec(f"from .{m} import *")
    except Exception as e:
        logging.getLogger("VLMP").error(e)
        logging.getLogger("VLMP").error(f"[Model] Error importing model type component {m}")



import sys, os

import copy

import logging

from . import modelBase
from ...utils.input import getLabelIndex

import pyGrained.models.AlphaCarbon as proteinModel

class SOP(modelBase):
    """
    Component name: SOP
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 24/03/2023

    Self organized polymer.

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"PDB",
                                                "centerInput",
                                                "SASA",
                                                "aggregateChains",
                                                "epsilonNC"},
                         requiredParameters  = {"PDB"},
                         definedSelections   = {"particleId","forceField"},
                         **params)

        ############################################################
        ######################  Set up model  ######################
        ############################################################

        #Check if PDB is a file or a PDB ID.
        #PDB is considered a file if it has a .pdb or .pqr extension.
        #If it is a file, it is loaded as a PDB file.
        #If it is a PDB ID, PDB file is downloaded from the RCSB PDB database.

        if params["PDB"].split(".")[-1] in ["pdb","pqr"]:
            inputPDBfilePath = params["PDB"]
        else:
            raise NotImplementedError("PDB ID download not implemented yet.")

        sopParams = {"SASA":params.get("SASA",False),
                      "centerInput":params.get("centerInput",True),
                      "aggregateChains":params.get("aggregateChains",True),
                      "parameters": copy.deepcopy(params)}

        sop = proteinModel.SelfOrganizedPolymer(name = name,
                                                inputPDBfilePath = inputPDBfilePath,
                                                params = sopParams)

        ########################################################

        types = self.getTypes()
        modelTypes = sop.getTypes()

        for _,t in modelTypes.items():
            types.addType(**t)

        #Set model
        self.setState(sop.getState())
        self.setStructure(sop.getStructure())
        self.setForceField(sop.getForceField())


    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += params["particleId"]

        if "forceField" in params:
            sel += self.getForceFieldSelection(params["forceField"])

        return sel



from VLMP.components.modelExtensions import modelExtensionBase

import numpy as np

class addBond(modelExtensionBase):

    """
    Component name: addBond
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire and Pablo Palacios
    Date: 19/06/2023

    Elastic Network Model (ENM).

    :param K: Spring constant.
    :type K: float
    :param r0: Equilibrium distance.
    :type r0: float

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"K","r0"},
                         requiredParameters  = {"K","r0"},
                         availableSelections = {"selection1","selection2"},
                         requiredSelections  = {"selection1","selection2"},
                         **params)

        K   = params["K"]
        r0  = params["r0"]

        ############################################################

        sel1Ids = self.getSelection("selection1")
        sel2Ids = self.getSelection("selection2")

        #Check that the selections have the same number of atoms, 1
        if len(sel1Ids) != 1 or len(sel2Ids) != 1:
            self.logger.error("Selections must have only one atom each.")
            raise ValueError("Selections must have only one atom each.")

        extension = {}

        extension[name] = {}
        extension[name]["type"]       = ["Bond2","Harmonic"]
        extension[name]["parameters"] = {}

        extension[name]["labels"] = ["id_i","id_j","K","r0"]
        extension[name]["data"]   = [[sel1Ids[0],sel2Ids[0],K,r0]]

        ############################################################

        self.setExtension(extension)



import os
import copy

import logging

################ TYPES INTERFACE ################

from pyUAMMD import simulation

class typesBase:

    def __init__(self,
                 _type:str,_name:str,
                 units,
                 availableParameters:set,
                 requiredParameters:set,
                 **params):

        self.logger = logging.getLogger("VLMP")

        self._type = _type
        self._name = _name

        self._units  = units

        self.availableParameters = availableParameters.copy()
        self.requiredParameters  = requiredParameters.copy()

        # Check all required parameters are available parameters
        if not self.requiredParameters.issubset(self.availableParameters):
            notAvailable = self.requiredParameters.difference(self.availableParameters)
            self.logger.error(f"[Types] ({self._type}) Some required parameters ({notAvailable}) are not available parameters for types {self._name}")
            raise Exception(f"Required paramaters are not available parameters")

        # Check if all parameters given by params are available
        for par in params:
            if par not in self.availableParameters:
                self.logger.error(f"[Types] ({self._type}) Parameter {par} not available for types {self._name}")
                raise Exception(f"Parameter not available")

        # Check if all required parameters are given
        for par in self.requiredParameters:
            if par not in params:
                self.logger.error(f"[Types] ({self._type}) Required parameter {par} not given for types {self._name}")
                raise Exception(f"Required parameter not given")

        self.logger.info(f"[Types] ({self._type}) Using types {self._name}")

        ########################################################

        self._typesUAMMD = None
        self._typesComp  = None
        self._typesDecl  = None

    ########################################################

    def getName(self):
        return self._name

    def getType(self):
        return self._type

    ########################################################

    def getUnits(self):
        return self._units

    ########################################################

    def getTypes(self):
        if self._typesComp is None:
            self.logger.error(f"[Types] ({self._type}) Types components not set for types {self._name}")
            raise Exception(f"Types components not set")
        if self._typesDecl is None:
            self.logger.error(f"[Types] ({self._type}) Types not declared for types {self._name}")
            raise Exception(f"Types not declared")

        types = {}

        components = [comp for comp in self.getTypesComponents().keys()]
        for typDecl in self._typesDecl:
            types[typDecl["name"]] = {comp:typDecl[comp] for comp in components}

        return copy.deepcopy(types)

    def setTypesName(self, typesUAMMD):
        self._typesUAMMD = typesUAMMD

    def getTypesName(self):
        if self._typesUAMMD is None:
            self.logger.error(f"[Types] ({self._type}) Types not set for types {self._name}")
            raise Exception(f"Types not set")
        return self._typesUAMMD

    def addTypesComponent(self, typesCompName, defaultValue):
        if self._typesComp is None:
            self._typesComp = {"name":None}
        if typesCompName in self._typesComp.keys():
            self.logger.error(f"[Types] ({self._type}) Component {typesCompName} already declared for types {self._name}.")
            raise Exception(f"Component already declared")
        self._typesComp[typesCompName] = defaultValue

    def getTypesComponents(self):
        if self._typesComp is None:
            self.logger.error(f"[Types] ({self._type}) Types components not set for types {self._name}")
            raise Exception(f"Types components not set")
        return self._typesComp

    def addType(self, **components):
        if self._typesComp is None:
            self.logger.error(f"[Types] ({self._type}) Types components not set for types {self._name}")
            raise Exception(f"Types components not set")
        if self._typesDecl is None:
            self._typesDecl = []
        #Check name in components
        if "name" not in components.keys():
            self.logger.error(f"[Types] ({self._type}) Name not given when declaring types {self._name}. Declaration: {components}")
            raise Exception(f"Name not given")

        #Check if all keys of components are available
        for key in components:
            if key not in self._typesComp.keys():
                self.logger.error(f"[Types] ({self._type}) Component {key} not available for types {self._name}. Available components are {list(self._typesComp.keys())}")
                raise Exception(f"Component not available")

        typeDecl = {}
        for key in self._typesComp.keys():
            if key in components:
                typeDecl[key] = components[key]
            else:
                typeDecl[key] = self._typesComp[key]
                self.logger.warning(f"[Types] ({self._type}) Component {key} not given for types {self._name}. Using default value {self._typesComp[key]}")

        addT = True
        for typ in self._typesDecl:
            if typ["name"] == typeDecl["name"]:
                addT = False
                #Check if all components are the same
                for key in self._typesComp.keys():
                    if typ[key] != typeDecl[key]:
                        typName = typ["name"]
                        self.logger.error(f"[Types] ({self._type}) Component {key} not the same for type \"{typName}\" (types {self._name})."\
                                          f" Previous value {typ[key]}, new value {typeDecl[key]}")
                        raise Exception(f"Component not the same")

        if addT:
            self.logger.debug(f"[Types] ({self._type}) Adding type {typeDecl['name']} for types {self._name}. Declaration: {typeDecl}")
            self._typesDecl.append(copy.deepcopy(typeDecl))

    ########################################################

    def getSimulation(self,DEBUG_MODE = False):

        labels = [comp for comp in self.getTypesComponents().keys()]

        data = []

        if self._typesDecl is None:
            self.logger.error(f"[Types] ({self._type}) Types not declared for types {self._name}")
            raise Exception(f"Types not declared")
        else:
            for types in self._typesDecl:
                data.append([types[comp] for comp in labels])

        return simulation({"global":{"types":{
                                     "type":["Types",self.getTypesName()],
                                     "labels":copy.deepcopy(labels),
                                     "data":copy.deepcopy(data)}}},DEBUG_MODE)

############### IMPORT ALL TYPES ###############

import glob

currentPath = os.path.dirname(os.path.abspath(__file__))
types = [os.path.basename(x) for x in glob.glob(currentPath+"/*.py")]
types = [x.split(".")[0] for x in types if "__" not in x]

for t in types:
    try:
        exec(f"from .{t} import *")
    except Exception as e:
        logging.getLogger("VLMP").error(e)
        logging.getLogger("VLMP").error(f"[Types] Error importing types type component {t}")



import sys, os

import logging

from . import simulationStepBase

class meanRadius(simulationStepBase):
    """
    Component name: meanRadius
    Component type: simulationStep

    Author: Pablo Ibáñez-Freire
    Date: 24/10/2023

    :param outputFilePath: Path to the output file
    :type outputFilePath: str

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"outputFilePath"},
                         requiredParameters  = {"outputFilePath"},
                         availableSelections = {"selection"},
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        parameters = {}

        parameters["outputFilePath"] = params.get("outputFilePath")

        simulationStep = {
            name:{
              "type":["GeometricalMeasure","MeanRadius"],
              "parameters":{**parameters}
            }
        }

        ############################################################

        self.setGroup("selection")
        self.setSimulationStep(simulationStep)






import sys, os

import logging

import numpy as np

from . import modelExtensionBase

class constantTorqueOverCenterOfMass(modelExtensionBase):

    """
    Component name: constantTorqueOverCenterOfMass
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 04/04/2023

    Constant torque over center of mass

    :param selection: Selection of particles where the force is applied
    :type selection: list of dictionaries
    :param torque: Torque applied to the center of mass
    :type torque: list of floats

    ...
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"torque"},
                         requiredParameters  = {"torque"},
                         availableSelections = {"selection"},
                         requiredSelections  = {"selection"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        torque = params.get("torque")
        # Check if the torque is a list of floats
        if not isinstance(torque,list):
            raise Exception("Torque must be a list of floats")

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["Set1","ConstantTorqueOverCenterOfMass"]
        extension[name]["parameters"] = {}
        extension[name]["labels"] = ["idSet_i","torque"]
        extension[name]["data"]   = []

        selectedIds = self.getSelection("selection")

        extension[name]["data"].append([selectedIds,torque])

        ############################################################

        self.setExtension(extension)






import os
import copy

import logging

################ ENSEMBLE INTERFACE ################

from pyUAMMD import simulation

from ...utils.input import getLabelIndex

class ensembleBase:

    def __init__(self,
                 _type:str,_name:str,
                 units,
                 types,
                 availableParameters:set,
                 requiredParameters:set,
                 **params):

        self.logger = logging.getLogger("VLMP")

        self._type = _type
        self._name = _name

        self._units = units
        self._types = types

        self.availableParameters = availableParameters.copy()
        self.requiredParameters  = requiredParameters.copy()

        # Check all required parameters are available parameters
        if not self.requiredParameters.issubset(self.availableParameters):
            notAvailable = self.requiredParameters.difference(self.availableParameters)
            self.logger.error(f"[Ensemble] ({self._type}) Some required parameters ({notAvailable}) \
                              are not available parameters for ensemble {self._name}")
            raise Exception(f"Required paramaters are not available parameters")

        # Check if all parameters given by params are available
        for par in params:
            if par not in self.availableParameters:
                self.logger.error(f"[Ensemble] ({self._type}) Parameter {par} \
                                    not available for ensemble {self._name}")
                raise Exception(f"Parameter not available")

        # Check if all required parameters are given
        for par in self.requiredParameters:
            if par not in params:
                self.logger.error(f"[Ensemble] ({self._type}) Required parameter \
                                    {par} not given for ensemble {self._name}")
                raise Exception(f"Required parameter not given")

        self.logger.info(f"[Ensemble] ({self._type}) Using ensemble {self._name}")

        ########################################################

        self._ensembleUAMMD = None
        self._ensembleComp  = None
        self._ensembleVal   = None

    ########################################################

    def getName(self):
        return self._name

    def getType(self):
        return self._type

    ########################################################

    def getUnits(self):
        return self._units

    def getTypes(self):
        return self._types

    ########################################################

    def setEnsembleName(self, ensembleUAMMD):
        self._ensembleUAMMD = ensembleUAMMD

    def getEnsembleName(self):
        if self._ensembleUAMMD is None:
            self.logger.error(f"[Ensemble] ({self._type}) Ensemble {self._name} has not been set")
            raise Exception(f"Ensemble not set")

        return self._ensembleUAMMD

    def addEnsembleComponent(self,componentName,componentValue):
        if self._ensembleComp is None:
            self._ensembleComp = {}
        if componentName in self._ensembleComp:
            self.logger.error(f"[Ensemble] ({self._type}) Component {componentName} \
                                already added to ensemble {self._name}")
            raise Exception(f"Component already added")

        self._ensembleComp[componentName] = componentValue

    def getEnsembleComponents(self):
        if self._ensembleComp is None:
            self.logger.error(f"[Ensemble] ({self._type}) No components added to ensemble {self._name}")
            raise Exception(f"No components added")

        return self._ensembleComp

    def getEnsembleComponent(self,componentName):
        if self._ensembleComp is None:
            self.logger.error(f"[Ensemble] ({self._type}) No components added to ensemble {self._name}")
            raise Exception(f"No components added")
        if componentName not in self._ensembleComp:
            self.logger.error(f"[Ensemble] ({self._type}) Component {componentName} \
                                not added to ensemble {self._name}")
            raise Exception(f"Component not added")

        return self._ensembleComp[componentName]

    ########################################################

    def getSimulation(self,DEBUG_MODE = False):
        labels = []
        data   = []

        for l in self.getEnsembleComponents():
            labels.append(l)
            data.append(self.getEnsembleComponent(l))

        return simulation({"global":{"ensemble":{
                                        "type":["Ensemble",self.getEnsembleName()],
                                        "labels":copy.deepcopy(labels),
                                        "data":[copy.deepcopy(data)]
                                                }
                                    }
                           },DEBUG_MODE)

############### IMPORT ALL ENSEMBLES ###############

import glob

currentPath = os.path.dirname(os.path.abspath(__file__))
ensembles_ = [os.path.basename(x) for x in glob.glob(currentPath+"/*.py")]
ensembles_ = [x.split(".")[0] for x in ensembles_ if "__" not in x]

for e in ensembles_:
    try:
        exec(f"from .{e} import *")
    except Exception as e:
        logging.getLogger("VLMP").error(e)
        logging.getLogger("VLMP").error(f"[Ensemble] Error importing ensemble type component \"{e}\"")



import sys, os
import copy

import logging

import numpy as np

from . import modelExtensionBase

class constraintParticlesPosition(modelExtensionBase):

    """
    Component name: constraintParticlesPosition
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 30/10/2023

    Apply a constraint to the position of a set of particles.

    :param selection: Selection of particles where the constraint will be applied
    :type selection: list of dictionaries
    :param K: Stiffness of the constraint
    :type K: float

    ...
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"K"},
                         requiredParameters  = {"K"},
                         availableSelections = {"selection"},
                         requiredSelections  = {"selection"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        K        = params.get("K")
        #Check if K is a float
        if not isinstance(K,float) and not isinstance(K,list):
            raise Exception("K must be a float or a list of floats")
        if isinstance(K,float):
            K = [K,K,K]

        r0 = [0.0,0.0,0.0]

        selectedIds  = self.getSelection("selection")
        idsPositions = self.getIdsState(selectedIds,"position")

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["Bond1","FixedHarmonic"]
        extension[name]["parameters"] = {}
        extension[name]["labels"] = ["id_i","position","K","r0"]
        extension[name]["data"]   = []
        for id_,pos in zip(selectedIds,idsPositions):
            extension[name]["data"].append([id_,pos,K,r0])

        ############################################################

        self.setExtension(extension)






import sys, os

import logging

from . import modelOperationBase

import numpy as np
from scipy.spatial import cKDTree

class setContactDistance(modelOperationBase):

    """
    Component name: setDistance
    Component type: modelOperation

    Author: Pablo Ibáñez-Freire
    Date: 28/08/2023

    Set the contact distance between two selections of particles.

    :param distance: distance to set
    :type distance: float
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"distance","resolution","inverse"},
                         requiredParameters  = {"distance","resolution"},
                         availableSelections = {"reference","mobile"},
                         requiredSelections  = {"reference","mobile"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        dst = params["distance"]
        res = params["resolution"]
        inv = params.get("inverse",False)

        referenceIds = self.getSelection("reference")
        mobileIds    = self.getSelection("mobile")

        referencePos  = np.asarray(self.getIdsState(referenceIds,"position"))
        referenceRads = np.asarray(self.getIdsProperty(referenceIds,"radius"))

        mobilePos  = np.asarray(self.getIdsState(mobileIds,"position"))
        mobileRads = np.asarray(self.getIdsProperty(mobileIds,"radius"))

        # Compute the total radius of each selection
        refCenter = np.mean(referencePos,axis=0)
        mobCenter = np.mean(mobilePos,axis=0)

        # Compute the vector between the centers
        centersVec = mobCenter-refCenter
        if inv:
            for i in range(mobilePos.shape[0]):
                mobilePos[i] = mobCenter-2.0*centersVec
            centersVec = -centersVec
        centersVec = centersVec/np.linalg.norm(centersVec)

        stepSize = res
        prevDist = None
        while True:

            minDst,minDstIndex = cKDTree(referencePos).query(mobilePos, 1)

            ############################################################

            mobileClosestIndex = np.argmin(minDst)
            mobileClosestDst   = minDst[mobileClosestIndex]

            mobileClosestPos = mobilePos[mobileClosestIndex]
            mobileClosestRad = mobileRads[mobileClosestIndex]

            ############################################################

            referenceClosestIndex = minDstIndex[mobileClosestIndex]
            referenceClosestDst   = minDst[mobileClosestIndex]

            referenceClosestPos = referencePos[referenceClosestIndex]
            referenceClosestRad = referenceRads[referenceClosestIndex]

            ############################################################

            currentDst = mobileClosestDst - (mobileClosestRad + referenceClosestRad)

            if np.abs(currentDst-dst) < res:
                break

            if prevDist is not None:
                if np.abs(currentDst-dst) > np.abs(prevDist-dst):
                    stepSize = stepSize/2

            if (currentDst < dst):
                for i in range(mobilePos.shape[0]):
                    mobilePos[i] = mobilePos[i] + centersVec * stepSize
                prevDist = currentDst

            if (currentDst > dst):
                for i in range(mobilePos.shape[0]):
                    mobilePos[i] = mobilePos[i] - centersVec * stepSize
                prevDist = currentDst

        newPos = []
        for i in range(len(mobileIds)):
            newPos.append(list(mobilePos[i]))

        self.setIdsState(mobileIds,"position",newPos)











import sys, os

import logging

from . import simulationStepBase

class stressMeasurement(simulationStepBase):
    """
    Component name: stressMeasurement
    Component type: simulationStep

    Author: Pablo Ibáñez-Freire
    Date: 29/09/2020

    This component writes the stress tensor of the system to a file.

    :param outputFilePath: Path to the output file
    :type outputFilePath: str
    :param radiusCutOff: Radius cutoff for the calculation of atom volumes
    :type radiusCutOff: float

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"outputFilePath","radiusCutOff"},
                         requiredParameters  = {"outputFilePath","radiusCutOff"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        parameters = {}

        parameters["outputFilePath"] = params["outputFilePath"]
        parameters["radiusCutOff"]   = params["radiusCutOff"]

        simulationStep = {
            name:{
              "type":["MechanicalMeasure","StressMeasure"],
              "parameters":{**parameters}
            }
        }

        ############################################################

        self.setSimulationStep(simulationStep)






from VLMP.components.modelExtensions import modelExtensionBase

import numpy as np
from scipy.optimize import fsolve

class surfaceMaxForce(modelExtensionBase):

    """
    Component name: surfaceMaxForce
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 17/06/2023

    Common epsilon, sigma surface for particles in the system.

    """

    def __LJtype2_maxForce_fit(self, sigma, radius, epsilon, maxForce):
        return (12.0*epsilon*((sigma/radius)**12 - (sigma/radius)**6)*(1.0/radius) - maxForce)**2

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"epsilon",
                                                "maxForce",
                                                "surfacePosition"},
                         requiredParameters  = {"surfacePosition",
                                                "maxForce"},
                         availableSelections = {"selection"},
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        surfacePosition = params.get("surfacePosition",0.0)
        epsilon         = params.get("epsilon",1.0)
        maxForce        = params.get("maxForce",1.0)

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["Surface","SurfaceGeneralLennardJonesType2"]
        extension[name]["parameters"] = {"surfacePosition":surfacePosition}

        extension[name]["labels"] = ["name","epsilon","sigma"]
        extension[name]["data"] = []

        types = self.getTypes()
        for typ,info in types.getTypes().items():
            radius = info["radius"]

            sgm = radius*np.power(maxForce*radius/(12.0*abs(epsilon)), 1/12) #Tentative sigma
            sgm,_,ier,msg = fsolve(self.__LJtype2_maxForce_fit, sgm, args=(radius, abs(epsilon), maxForce), full_output=True)

            if ier != 1:

                maxForceSteps = np.linspace(maxForce/4.0, maxForce, 10)
                for mFS in maxForceSteps:
                    sgm = radius*np.power(mFS*radius/(12.0*abs(epsilon)), 1/12)
                    sgm,_,ier,msg = fsolve(self.__LJtype2_maxForce_fit, sgm, args=(radius, abs(epsilon), mFS), full_output=True)

                    if ier == 1:
                        break

                if ier != 1:
                    self.logger.error(f"[ModelExtension] (surfaceMaxForce) Could not find sigma for type {typ}. Message: {msg}")
                    raise Exception("Error in sigma calculation")

            sgm = sgm[0]
            self.logger.debug(f"[ModelExtension] (surfaceMaxForce) Sigma for type {typ} with radius {radius} is {sgm}")

            extension[name]["data"].append([typ,epsilon,round(sgm,2)])

        ############################################################

        self.setExtension(extension)



import sys, os

import logging

from . import modelOperationBase

import numpy as np

class setCenterOfMassPosition(modelOperationBase):

    """
    Component name: setCenterOfMassPosition
    Component type: modelOperation

    Author: Pablo Ibáñez-Freire
    Date: 17/06/2023

    Set the center of mass of a selection of particles to a given position.

    :param position: Position to set the center of mass to.
    :type position: list of floats

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"position"},
                         requiredParameters  = {"position"},
                         availableSelections = {"selection"},
                         requiredSelections  = {"selection"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        selectedIds = self.getSelection("selection")

        masses = np.asarray(self.getIdsProperty(selectedIds,"mass"))
        pos    = np.asarray(self.getIdsState(selectedIds,"position"))

        totalMass = np.sum(masses)
        com       = np.sum(masses[:,np.newaxis]*pos,axis=0)/totalMass

        translation = np.asarray(params.get("position")) - com

        pos += translation
        pos = pos.tolist()

        self.setIdsState(selectedIds,"position",pos)

